// Om Ganesayanamaha
package org.processmining.plugins.tracealignment.clustering;

/**
 * 
 * @author R.P. Jagadeesh Chandra Bose (JC)
 * @date 19 Dec 2009
 * @version 0.1
 * @email j.c.b.rantham.prabhakara@tue.nl
 */
public class AgglomerativeHierarchicalClustering {
	public class Cluster {
		int cardinality; // the number of elements in this cluster
		Cluster left, right; // the left and right child (cluster) of this node
		float[] clusterDistanceSimilarityArray; // the distance/similarity of this cluster with other clusters

		public Cluster(float[] clusterDistanceSimilarityArray) {
			cardinality = 1;
			this.clusterDistanceSimilarityArray = clusterDistanceSimilarityArray;
			left = null;
			right = null;
		}

		public Cluster(Cluster left, Cluster right, float[] clusterDistanceSimilarityArray) {
			this.left = left;
			this.right = right;
			cardinality = left.cardinality + right.cardinality;
			this.clusterDistanceSimilarityArray = clusterDistanceSimilarityArray;
		}

		public boolean live() {
			return clusterDistanceSimilarityArray != null;
		}

		public void kill() {
			clusterDistanceSimilarityArray = null;
		}
	}

	int K; // number of clusters created so far
	int N; // the number of items in the data set (for clustering)
	Cluster[] clusters; // the nodes in the hierarchy

	int[][] itemsJoined; // the items (can be individual elements or clusters already formed) joined (as a cluster) in an iteration

	public enum ClusterCriteria {
		MIN_VARIANCE, SINGLE_LINKAGE, COMPLETE_LINKAGE, AVERAGE_LINKAGE, CENTROID
	};

	public enum ClusterType {
		DISTANCE, SIMILARITY
	};

	ClusterCriteria clusterCriteria;
	ClusterType clusterType;

	public AgglomerativeHierarchicalClustering(float[][] distanceSimilarityMatrix, ClusterType clusterType,
			ClusterCriteria clusterCriteria) {
		this.clusterType = clusterType;
		this.clusterCriteria = clusterCriteria;

		N = distanceSimilarityMatrix.length;

		itemsJoined = new int[N - 1][2]; // since there are n items, the tree would contain n leaves and n-1 inner nodes (total 2n-1 nodes)
		clusters = new Cluster[2 * N - 1];

		// Create the leaf nodes (clusters)

		for (int i = 0; i < N; i++) {
			clusters[i] = new Cluster(distanceSimilarityMatrix[i]);
		}

		K = N;
		while (K < 2 * N - 1) {
			findAndJoin();
		}
	}

	private void findAndJoin() {
		// the index (i,j) for the maximum similarity nodes
		int minMaxI = -1, minMaxJ = -1;
		float maxSimilarity = Float.NEGATIVE_INFINITY;
		float minDistance = Float.POSITIVE_INFINITY;
		float s, d;

		if (clusterType.equals(ClusterType.SIMILARITY)) {
			// scan through all the nodes formed so far and pick the two with maximum similarity
			for (int i = 0; i < K; i++) {
				if (clusters[i].live()) {
					for (int j = 0; j < i; j++) {
						if (clusters[j].live()) {
							s = distSim(i, j);
							if (s > maxSimilarity) {
								minMaxI = i;
								minMaxJ = j;
								maxSimilarity = s;
							}
						}
					}
				}
			}
		} else {
			// scan through all the nodes formed so far and pick the two with minimum distance
			for (int i = 0; i < K; i++) {
				if (clusters[i].live()) {
					for (int j = 0; j < i; j++) {
						if (clusters[j].live()) {
							d = distSim(i, j);
							if (d < minDistance) {
								minMaxI = i;
								minMaxJ = j;
								minDistance = d;
							}
						}
					}
				}
			}
		}

		itemsJoined[K - N][0] = minMaxI;
		itemsJoined[K - N][1] = minMaxJ;

		switch (clusterCriteria) {
			case MIN_VARIANCE :
				joinMinVariance(minMaxI, minMaxJ);
				break;
			case SINGLE_LINKAGE :
				joinSingleLinkage(minMaxI, minMaxJ);
				break;
			case COMPLETE_LINKAGE :
				joinCompleteLinkage(minMaxI, minMaxJ);
				break;
			case AVERAGE_LINKAGE :
				joinAverageLinkage(minMaxI, minMaxJ);
				break;
			case CENTROID :
				joinCentroid(minMaxI, minMaxJ);
				break;
		}
	}

	private float distSim(int i, int j) {
		return clusters[Math.max(i, j)].clusterDistanceSimilarityArray[Math.min(i, j)];
	}

	private void joinMinVariance(int i, int j) {
		// join clusters i and j to form a new cluster K
		float[] clusterDistanceSimilarityArray = new float[K];
		for (int m = 0; m < K; m++) {
			if (clusters[m].live() && (m != j) && (m != i)) {
				clusterDistanceSimilarityArray[m] = ((float) (clusters[i].cardinality + clusters[m].cardinality))
						/ (clusters[i].cardinality + clusters[j].cardinality + clusters[m].cardinality) * distSim(m, i)
						+ ((float) (clusters[j].cardinality + clusters[m].cardinality))
						/ (clusters[i].cardinality + clusters[j].cardinality + clusters[m].cardinality) * distSim(m, j)
						- ((float) (clusters[m].cardinality))
						/ (clusters[i].cardinality + clusters[j].cardinality + clusters[m].cardinality) * distSim(i, j);
			}
		}
		clusters[K] = new Cluster(clusters[i], clusters[j], clusterDistanceSimilarityArray);
		clusters[i].kill();
		clusters[j].kill();
		K++;
	}

	private void joinAverageLinkage(int i, int j) {
		// Join clusters i and j to form a new cluster K
		float[] clusterDistanceSimilarityArray = new float[K];

		for (int m = 0; m < K; m++) {
			if (clusters[m].live() && (m != j) && (m != i)) {
				clusterDistanceSimilarityArray[m] = (distSim(i, m) * clusters[i].cardinality + distSim(j, m)
						* clusters[j].cardinality)
						/ (clusters[i].cardinality + clusters[j].cardinality);
			}
		}
		clusters[K] = new Cluster(clusters[i], clusters[j], clusterDistanceSimilarityArray);
		clusters[i].kill();
		clusters[j].kill();
		K++;
	}

	private void joinSingleLinkage(int i, int j) {
		// Join clusters i and j to form a new cluster K
		float[] clusterDistanceSimilarityArray = new float[K];

		for (int m = 0; m < K; m++) {
			if (clusters[m].live() && (m != j) && (m != i)) {
				clusterDistanceSimilarityArray[m] = Math.min(distSim(i, m), distSim(j, m));
			}
		}
		clusters[K] = new Cluster(clusters[i], clusters[j], clusterDistanceSimilarityArray);
		clusters[i].kill();
		clusters[j].kill();
		K++;
	}

	private void joinCompleteLinkage(int i, int j) {
		// Join clusters i and j to form a new cluster K
		float[] clusterDistanceSimilarityArray = new float[K];

		for (int m = 0; m < K; m++) {
			if (clusters[m].live() && (m != j) && (m != i)) {
				clusterDistanceSimilarityArray[m] = Math.max(distSim(i, m), distSim(j, m));
			}
		}
		clusters[K] = new Cluster(clusters[i], clusters[j], clusterDistanceSimilarityArray);
		clusters[i].kill();
		clusters[j].kill();
		K++;
	}

	private void joinCentroid(int i, int j) {
		// Join clusters i and j to form a new cluster K
		float[] clusterDistanceSimilarityArray = new float[K];

		for (int m = 0; m < K; m++) {
			if (clusters[m].live() && (m != j) && (m != i)) {
				clusterDistanceSimilarityArray[m] = (clusters[i].cardinality * distSim(m, i) + clusters[j].cardinality
						* distSim(m, j) - ((clusters[i].cardinality * clusters[j].cardinality * distSim(i, j)) / ((clusters[i].cardinality + clusters[j].cardinality))))
						/ (clusters[i].cardinality + clusters[j].cardinality);
			}
		}
		clusters[K] = new Cluster(clusters[i], clusters[j], clusterDistanceSimilarityArray);
		clusters[i].kill();
		clusters[j].kill();
		K++;
	}

	public Cluster[] getClusters() {
		return clusters;
	}

	public int[][] getItemsJoined() {
		return itemsJoined;
	}

}
