// Om Ganesayanamaha
package org.processmining.plugins.tracealignment.ui;

import java.awt.Color;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

@SuppressWarnings( { "unused", "cast" })
public class Alignment implements SequenceListener, Comparator<Sequence> {
	// The encoding Length of the activities in the log
	public int encodingLength = 1;

	// The list of sequences defining the alignment
	private final Sequence[] sequencesInAlignmentArray;

	// The reference sequence for the alignment
	private Sequence referenceSequence = null;

	// Flag to indicate whether any changes have been done to the alignment

	private boolean alignmentChanged = false;

	// Font size for displaying the sequences; Default is set to 12 
	private final int fontSize = 16;

	// List of listeners that do the required things when something is changed in the alignment
	private final List<AlignmentListener> listenersList = new ArrayList<AlignmentListener>();

	// Alignment name

	private String name;

	private int lastRulerAnnIndexSelected = -1;
	private int lastColumnIndexSelected = -1;

	private int columnFilterCurrColumn = -1;
	Map<Integer, ColumnFilter> columnFilterMap = new HashMap<Integer, ColumnFilter>();

	private int columnSortCurrColumn = -1;
	Map<Integer, ColumnSort> columnSortMap = new HashMap<Integer, ColumnSort>();

	boolean columnSorting = true;
	boolean columnFiltering = false;

	private final List<RulerAnnotation> rulerAnnotationList = new ArrayList<RulerAnnotation>();
	private final ArrayList<Boolean> columnsSelectedList = new ArrayList<Boolean>();

	private int lastActivityIndexSelected = -1;
	private int lastActivityRowSelected = -1;

	public Alignment(int encodingLength, Sequence[] sequencesInAlignmentArray) {
		this.encodingLength = encodingLength;
		this.sequencesInAlignmentArray = sequencesInAlignmentArray;
		// add the sequence listener implemented here for the sequences
		for (Sequence s : sequencesInAlignmentArray) {
			s.addListener(this);
		}
	}

	// add/remove listeners
	public void addListener(AlignmentListener listener) {
		listenersList.add(listener);
	}

	public String getName() {
		return name;
	}

	/**
	 * @return the number of sequences defined in this alignment
	 */
	public int getNoSequences() {
		return sequencesInAlignmentArray.length;
	}

	public int getFontSize() {
		return fontSize;
	}

	/**
	 * 
	 * @param index
	 * @return a specific sequence in the alignment identified by the index or
	 *         null in case the index is out of range or the sequence doesn't
	 *         exist
	 */
	public Sequence getSequence(int index) {
		if ((index < 0) || (index >= sequencesInAlignmentArray.length)) {
			return null;
		}

		if (sequencesInAlignmentArray[index] != null) {
			return sequencesInAlignmentArray[index];
		} else {
			return null;
		}
	}

	public Sequence getSequence(String sequenceName) {
		for (Sequence sequence : sequencesInAlignmentArray) {
			if (sequence.getName().equals(sequenceName)) {
				return sequence;
			}
		}
		return null;
		//TODO changed from original
		//        return (Sequence) sequence_name_map.get(seq_name);
	}

	// maximum length, computed if necessary
	public int maxLength() {
		int maxLength = 0;
		for (int i = sequencesInAlignmentArray.length - 1; i >= 0; i--) {
			int length = sequencesInAlignmentArray[i].getLengthWithGaps();
			if (length > maxLength) {
				maxLength = length;
			}
		}
		return maxLength;
	}

	public boolean isColumnSelected() {
		if (columnsSelectedList.isEmpty()) {
			return false;
		}
		return columnsSelectedList.contains(new Boolean(true));
	}

	/**
	 * Returns the total number of columns in the alignment
	 * 
	 * @return
	 */
	public int getColumnsSize() {
		return columnsSelectedList.size();
	}

	public void addColumn() {
		columnsSelectedList.add(new Boolean(false));
	}

	public void removeColumnFromEnd() {
		columnsSelectedList.remove(columnsSelectedList.size() - 1);
	}

	public int getFirstColumnSelected() {
		return columnsSelectedList.indexOf(new Boolean(true));
	}

	public int getLastColumnSelected() {
		return columnsSelectedList.lastIndexOf(new Boolean(true));
	}

	public int getNumberOfColumnsSelected() {
		int noSelectedColumns = 0;
		for (int i = 0; i < columnsSelectedList.size(); i++) {
			if (isColumnSelected(i)) {
				noSelectedColumns++;
			}
		}
		return noSelectedColumns;
	}

	public int getLastColumnIdxSelected() {
		return lastColumnIndexSelected;
	}

	public void setLastColumnIdxSelected(int idx) {
		lastColumnIndexSelected = idx;
		clearLastActivityIndexSelected();
	}

	public boolean isColumnSelected(int idx) {
		if (columnsSelectedList.isEmpty()) {
			return false;
		}
		return columnsSelectedList.get(idx).booleanValue();
	}

	public void removeColumnRange(int from, int to) {
		for (int i = 0; i < to - from; i++) {
			columnsSelectedList.remove(from);
		}
	}

	public void setColumnFilterCurrentColumn(int col) {
		columnFilterCurrColumn = col;
	}

	public int getColumnFilterCurrentColumn() {
		return columnFilterCurrColumn;
	}

	public int getColumnSortCurrentColumn() {
		return columnSortCurrColumn;
	}

	public void setColumnSortCurrentColumn(int col) {
		columnSortCurrColumn = col;
	}

	public ColumnSort getCurrentColumnSort() {
		return columnSortMap.get(columnSortCurrColumn);
	}

	public boolean getColumnSorting() {
		return columnSorting;
	}

	public void setColumnSorting(boolean b) {
		columnSorting = b;
	}

	public boolean getColumnFiltering() {
		return columnFiltering;
	}

	public void setColumnFiltering(boolean b) {
		columnFiltering = b;
	}

	public ColumnSort getColumnSort(Integer key) {
		return columnSortMap.get(key);
	}

	public void addColumnSort(ColumnSort sort, int column) {
		ColumnSort old = columnSortMap.get(column);
		if (old != null) {
			((ColumnSort) old).removed();
			columnSortMap.remove(column);
		}
		columnSortMap.put(column, sort);
	}

	public void removeColumnSort(int column) {
		ColumnSort columnSort = columnSortMap.get(column);
		if (columnSort != null) {
			//TODO removed the invocation of removed() on column sort; mostly dealing with priority settings
			columnSort.removed();
			columnSortMap.remove(column);
		}
	}

	public void sortBasedOnColumnSorts() throws Exception {
		// score all sequences based on column sorts
		calcSortValuesBasedOnColumnSorts();

		// sort sequence components based on the sortValue setup above
		sortBasedOnSortValue();
	}

	public void sortBasedOnSortValue() throws Exception {
		AlignmentSorter sorter = new AlignmentSorter(this, this);
		sorter.sort();
	}

	void calcSortValuesBasedOnColumnSorts() {
		// now for each sequence...
		for (int i = 0; i < sequencesInAlignmentArray.length; i++) {
			// calculate the sort value based on all column sorts
			calcSortValueBasedOnColumnSorts(sequencesInAlignmentArray[i]);
		}
	}

	public void calcSortValueBasedOnColumnSorts(Sequence seq) {
		// start with zero
		int sortValue = 0;

		// add in bits based on column sorts
		java.util.Iterator<Integer> iter = getColumnSortColumnIterator();
		while (iter.hasNext()) {
			Integer column = iter.next();
			ColumnSort cs = columnSortMap.get(column);

			// get activity at this column
			String encodedActivity = seq.getEncodedActivity(column);

			// add in this bit
			sortValue |= cs.optionTestGetPriorityBit(encodedActivity);
		}

		// set this
		seq.setSortValue(sortValue);
	}

	public java.util.Iterator<Integer> getColumnSortColumnIterator() {
		return columnSortMap.keySet().iterator();
	}

	public int compare(Sequence seq1, Sequence seq2) {
		return seq2.getSortValue() - seq1.getSortValue();
	}

	public String[] getUniqueActivitiesInColumn(int column) {
		Set<String> encodedActivitiesInColumnSet = new HashSet<String>();
		Sequence[] seqs = getAllSequences();

		for (int i = 0; i < seqs.length; i++) {
			String encodedActivity = seqs[i].getEncodedActivity(column);
			if (encodedActivity.equals("") || encodedActivity.contains("-") || encodedActivity.contains(" ")) {
				continue;
			}
			encodedActivitiesInColumnSet.add(encodedActivity);
		}

		String[] encodedActivitiesInColumnArray = new String[encodedActivitiesInColumnSet.size()];
		int index = 0;
		for (String encodedActivity : encodedActivitiesInColumnSet) {
			encodedActivitiesInColumnArray[index++] = encodedActivity;
		}

		return encodedActivitiesInColumnArray;
	}

	public int[] getUniqueActivityCounts(int column, String[] encodedActivitiesArray) {
		Sequence[] seqs = getAllSequences();
		return getUniqueActivityCounts(column, encodedActivitiesArray, seqs);
	}

	public int[] getUniqueActivityCounts(int column, String[] encodedActivitiesArray, Sequence[] seqs) {
		int[] counts = new int[encodedActivitiesArray.length];
		for (int i = 0; i < counts.length; i++) {
			int count = 0;
			String encodedActivity = encodedActivitiesArray[i];
			for (int j = 0; j < seqs.length; j++) {
				if (!seqs[j].getEncodedActivity(column).equals(encodedActivity)) {
					continue;
				}
				count++;
			}
			counts[i] = count;
		}

		return counts;
	}

	public void addColumnFilter(ColumnFilter filter, int column) {
		ColumnFilter columnFilter = columnFilterMap.get(column);
		if (columnFilter != null) {
			columnFilterMap.remove(column);
		}
		columnFilterMap.put(column, filter);
	}

	public void removeColumnFilter(int column) {
		ColumnFilter columnFilter = columnFilterMap.get(column);
		if (columnFilter != null) {
			columnFilterMap.remove(column);
		}
	}

	public ColumnFilter getCurrentColumnFilter() {
		return columnFilterMap.get(columnFilterCurrColumn);
	}

	public ColumnFilter getColumnFilter(Integer key) {
		return columnFilterMap.get(key);
	}

	public java.util.Iterator<Integer> getColumnFilterColumnIterator() {
		return columnFilterMap.keySet().iterator();
	}

	public boolean isRulerAnnotationSelected() {
		boolean isRulerAnnotationSelected = false;

		for (RulerAnnotation rulerAnnotation : rulerAnnotationList) {
			if (rulerAnnotation.isSelected()) {
				isRulerAnnotationSelected = true;
			}
		}

		return isRulerAnnotationSelected;
	}

	public int rulerAnnotationsSize() {
		return rulerAnnotationList.size();
	}

	public void addRulerAnnotation() {
		rulerAnnotationList.add(new RulerAnnotation(null, Color.blue, false));
	}

	public void removeRulerAnnotationFromEnd() {
		rulerAnnotationList.remove(rulerAnnotationList.size() - 1);
	}

	public void removeRulerAnnotationRange(int from, int to) {
		for (int i = 0; i < to - from; i++) {
			rulerAnnotationList.remove(from);
		}
	}

	public RulerAnnotation getRulerAnnotation(int idx) {
		if ((idx != -1) && (idx < rulerAnnotationList.size())) {
			return rulerAnnotationList.get(idx);
		} else {
			return (new RulerAnnotation(null, Color.white, false));
		}
	}

	public void setLastRulerAnnIdxSelected(int idx) {
		lastRulerAnnIndexSelected = idx;
		clearLastActivityIndexSelected();
	}

	/**
	 * @return the list (array) of all sequences defined by this alignment
	 */
	public Sequence[] getAllSequences() {
		Sequence[] tempSequenceArray = new Sequence[sequencesInAlignmentArray.length];
		System.arraycopy(sequencesInAlignmentArray, 0, tempSequenceArray, 0, sequencesInAlignmentArray.length);
		return tempSequenceArray;
	}

	// find the index of a sequence
	public int getIndex(Sequence s) {
		//TODO changed from the original; original uses a HashMap
		int noSequencesInAlignment = getNoSequences();
		for (int i = 0; i < noSequencesInAlignment; i++) {
			if (sequencesInAlignmentArray[i] == s) {
				return i;
			}
		}
		return -1;
	}

	// swap two sequences
	public void swapSequence(int i, int j) throws Exception {
		if ((i < 0) || (i >= sequencesInAlignmentArray.length) || (j < 0) || (j >= sequencesInAlignmentArray.length)) {
			throw new Exception("Bad indexes for sequence swap");
		}
		if (i == j) {
			return;
		}

		Sequence s = sequencesInAlignmentArray[i];
		sequencesInAlignmentArray[i] = sequencesInAlignmentArray[j];
		sequencesInAlignmentArray[j] = s;

		setChanged(true);
		for (AlignmentListener alignmentListener : listenersList) {
			alignmentListener.alignmentSeqSwapped(this, i, j);
		}
	}

	public void setChanged(boolean b) {
		alignmentChanged = b;
	}

	public Sequence getRefSequence() {
		return referenceSequence;
	}

	public void clearLastActivityIndexSelected() {
		lastActivityIndexSelected = -1;
		lastActivityRowSelected = -1;
	}

	public void setLastActivityIndexSelected(int row, int idx) {
		lastActivityRowSelected = row;
		lastActivityIndexSelected = idx;
	}

	public int getLastActivityIndexSelected() {
		return lastActivityIndexSelected;
	}

	public int getLastActivityRowSelected() {
		return lastActivityRowSelected;
	}

	public void clearLastColumnIndexSelected() {
		lastColumnIndexSelected = -1;
	}

	public void clearLastRulerAnnIndexSelected() {
		lastRulerAnnIndexSelected = -1;
	}

	public void deselectAllRulerSelections() {
		for (RulerAnnotation rulerAnnotation : rulerAnnotationList) {
			rulerAnnotation.setSelected(false);
		}

		lastRulerAnnIndexSelected = -1;
	}

	public void deselectAllColumns(DisplayProperties p) {
		deselectAllColumns(p, true);
	}

	public void deselectAllColumns(DisplayProperties p, boolean quickDeselect) {
		for (int idx = 0; idx < columnsSelectedList.size(); idx++) {
			if (quickDeselect) {
				columnsSelectedList.set(idx, new Boolean(false));
			} else {
				setColumnSelected(idx, false, p);
			}
		}
		lastColumnIndexSelected = -1;
	}

	public void setColumnSelected(int idx, boolean b, DisplayProperties p) {
		if (idx < columnsSelectedList.size()) {
			columnsSelectedList.set(idx, new Boolean(b));
			int noSequencesInAlignment = getNoSequences();
			for (int row = 0; row < noSequencesInAlignment; row++) {
				p.setSeqHighlight(this.getSequence(row), idx, b, false);
			}
			p.fireDisplayHighlightsChanged(getAllSequences());
		}
	}

	// after calling this method the caller must ensure that the
	// GUI is updated...in this case its the part
	// that displays the currently selected column
	// and which ref col # that is...
	public void setRefSequence(Sequence sequence) {
		if (referenceSequence == null) {
			if (sequence == null) {
				return;
			}
		} else {
			if (referenceSequence.equals(sequence)) {
				return;
			}
		}

		referenceSequence = sequence;
	}

	public void removeAllGapColumns() {
		Sequence sequence;
		List<Integer> allGapColumnList = getGapColumnList();
		List<Integer> rangeCols = new ArrayList<Integer>();
		int left, right, column, neighbors;

		//        for (int col = 0; col <= alignment.getLastColumnSelected(); col++) {
		//			if (alignment.isColumnSelected(col)) {
		//				selectedColumnList.add(new Integer(col));
		//			}
		//		}

		for (int i = 0; i < allGapColumnList.size(); i++) {
			neighbors = 0;
			column = ((Integer) allGapColumnList.get(i)).intValue();
			left = i == 0 ? column : ((Integer) allGapColumnList.get(i - 1)).intValue();
			right = i == allGapColumnList.size() - 1 ? column : ((Integer) allGapColumnList.get(i + 1)).intValue();
			if (column - left == 1) {
				neighbors++;
			}
			if (right - column == 1) {
				neighbors++;
			}
			if (neighbors == 0) {
				rangeCols.add(new Integer(column));
				rangeCols.add(new Integer(column));
			}
			if (neighbors == 1) {
				rangeCols.add(new Integer(column));
			}
		}
		int start, end;

		for (int row = 0; row < getNoSequences(); row++) {
			sequence = getSequence(row);
			for (int i = rangeCols.size() - 1; i >= 1; i -= 2) {
				start = ((Integer) rangeCols.get(i)).intValue();
				end = ((Integer) rangeCols.get(i - 1)).intValue();
				try {
					if (row == 0) {
						removeColumnRange(end, start + 1);
						removeRulerAnnotationRange(end, start + 1);
					}
					sequence.shiftActivity(start + 1, end - start - 1, true);
				} catch (Exception exp) {
					exp.printStackTrace();
				}
			}
		}
	}

	private List<Integer> getGapColumnList() {
		List<Integer> gapColumnList = new ArrayList<Integer>();
		int alignmentLength = maxLength();
		int noSequencesInAlignment = getNoSequences();
		String encodedActivity;
		boolean activityFound;
		for (int i = 0; i < alignmentLength; i++) {
			activityFound = false;
			for (int j = 0; j < noSequencesInAlignment; j++) {
				encodedActivity = sequencesInAlignmentArray[j].getEncodedActivity(i);
				if (!encodedActivity.contains("-")) {
					activityFound = true;
					break;
				}
			}
			if (!activityFound) {
				gapColumnList.add(i);
			}
		}
		return gapColumnList;
	}

	// Ruler Annotation class
	public static class RulerAnnotation {
		private String letter;
		private Color color;
		private boolean selected;

		public RulerAnnotation(String letter, Color color, boolean selected) {
			this.letter = letter;
			this.color = color;
			this.selected = selected;
		}

		public String getLetter() {
			return letter;
		}

		public Color getColor() {
			return color;
		}

		public boolean isSelected() {
			return selected;
		}

		public void setLetter(String newLetter) {
			letter = newLetter;
		}

		public void setColor(Color newColor) {
			color = newColor;
		}

		public void setSelected(boolean newSelected) {
			selected = newSelected;
		}
	}

	//Seuence Listener interface implementation
	public void sequenceNameChanged(Sequence sequence, String oldName) throws Exception {
		setChanged(true);
		//TODO original had a little bit more
	}

	public void sequenceActivityChanged(Sequence sequence) {
		setChanged(true);
		for (AlignmentListener listener : listenersList) {
			listener.alignmentSeqActivityChanged(this, sequence);
		}
	}

	public void sequenceAnnotationChanged(Sequence sequence) {
		setChanged(true);
	}

	public void sequenceColorChanged(Sequence sequence) {
		// Nothing to be done here

	}

	public void sequenceColumnAnnotationsChanged(Sequence sequence, int column) {
		// Nothing to be done here
	}

	public void sequenceLineAnnotationsChanged(Sequence aaseq) {
		// Nothing to be done here
	}
	
	public int getEncodingLength(){
		return encodingLength;
	}
}
