package org.processmining.plugins.tracealignment.ui;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;

import javax.swing.BoxLayout;
import javax.swing.JComponent;
import javax.swing.JPanel;
import javax.swing.Scrollable;
import javax.swing.SwingConstants;

import org.processmining.plugins.tracealignment.AlignmentFrame;

@SuppressWarnings("serial")
public class AlignmentPanel extends JPanel implements MouseListener, MouseMotionListener, MouseWheelListener,
		KeyListener, AlignmentListener, Scrollable {

	protected Alignment alignment;
	protected DisplayProperties displayProperties;
	protected AlignmentFrame owner;

	public AlignmentPanel(AlignmentFrame owner, Alignment alignment, DisplayProperties displayProperties) {
		super();

		setBackground(Color.white);
		setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));

		this.owner = owner;
		this.alignment = alignment;
		this.displayProperties = displayProperties;

		alignment.addListener(this);

		int noSequencesInAlignment = alignment.getNoSequences();

		for (int i = 0; i < noSequencesInAlignment; i++) {
			Sequence currentSequence = alignment.getSequence(i);
			JComponent sequenceComponent = new SequenceComponent(owner, currentSequence, displayProperties);
			sequenceComponent.setLayout(new BoxLayout(sequenceComponent, BoxLayout.Y_AXIS));
			this.add(sequenceComponent);
		}

		addMouseListener(this);
		addMouseMotionListener(this);
		addMouseWheelListener(this);
	}

	public boolean hasCollapsedRows() {
		Component[] components = getComponents();
		for (int i = 0; i < components.length; i++) {
			Component component = components[i];
			if (component instanceof SequenceComponent) {
				SequenceComponent sc = (SequenceComponent) component;
				if (sc.collapsed) {
					return true;
				}
			}
		}
		return false;
	}

	/**
	 * Determine the actual row given visible row index; When sequences are
	 * hidden, the cursor based row index (visible row) need to be adjusted for
	 * the actual sequence in order to perform any operations on the sequence
	 * 
	 * @param visibleRow
	 * @return the actual row index of the sequence in the alignment
	 */
	public int getRowIndex(int visibleRow) {
		Component[] components = getComponents();
		int rowIndex = 0;
		for (int i = 0; i < components.length; i++) {
			Component component = components[i];
			if (component instanceof SequenceComponent) {
				SequenceComponent sc = (SequenceComponent) component;
				if (sc.collapsed) {
					continue;
				}
			}

			if (rowIndex == visibleRow) {
				return i;
			}

			rowIndex++;
		}
		return -1;
	}

	public int getNumVisibleRows() {
		int numVisibleRows = 0;

		// count the # of uncollapsed rows from 0 to real row
		Component[] components = getComponents();
		for (int i = 0; i < components.length; i++) {
			Component component = components[i];
			if (component instanceof SequenceComponent) {
				SequenceComponent sc = (SequenceComponent) component;
				if (sc.collapsed) {
					continue;
				}
			}
			numVisibleRows++;
		}
		return numVisibleRows;
	}

	// get visible row given real row
	public int getInverseRowIndex(int realRow) {
		int invRow = 0;

		// count the # of uncollapsed rows from 0 to real row
		Component[] components = getComponents();
		for (int i = 0; i < components.length; i++) {
			if (realRow == i) {
				return invRow;
			}

			Component component = components[i];
			if (component instanceof SequenceComponent) {
				SequenceComponent sc = (SequenceComponent) component;
				if (sc.collapsed) {
					continue;
				}
			}
			invRow++;
		}
		return -1;
	}

	public int getSequenceIndex(Sequence sequence) {
		Component[] components = getComponents();
		for (int i = 0; i < components.length; i++) {
			Component component = components[i];
			if (component instanceof SequenceComponent) {
				SequenceComponent sc = (SequenceComponent) component;
				if (sc.getSequence().equals(sequence)) {
					return i;
				}
			}
		}

		return -1;
	}

	public SequenceComponent getSequenceComponent(int index) {
		Component[] components = getComponents();
		if (index < components.length) {
			return (SequenceComponent) components[index];
		} else {
			return null;
		}
	}

	public SequenceComponent getSequenceComponent(Sequence sequence) {
		int index = getSequenceIndex(sequence);
		if (index < 0) {
			return null;
		}
		return getSequenceComponent(index);
	}

	public void setAllSequenceCollapsed(boolean collapsed, boolean repaint) {
		Component[] components = getComponents();
		for (int i = 0; i < components.length; i++) {
			Component component = components[i];
			if (component instanceof SequenceComponent) {
				SequenceComponent sc = (SequenceComponent) component;
				sc.setCollapsed(collapsed);
			}
		}

		if (repaint) {
			revalidate();
			repaint();
		}
	}

	public void setSequenceCollapsed(int idx, boolean collapsed, boolean repaint) {
		SequenceComponent sc = this.getSequenceComponent(idx);
		sc.setCollapsed(collapsed);

		if (repaint) {
			revalidate();
			repaint();
		}
	}

	public Dimension getPreferredSize() {
		return getLayout().preferredLayoutSize(this);
	}

	private int findRow(int x, int y) {
		Component c = findComponentAt(x, y);
		Component[] components = getComponents();
		int index;

		for (index = components.length - 1; index >= 1; index--) {
			if (components[index] == c) {
				break;
			}
		}
		return index;
	}

	// Tooltips for column annotations
	public String getToolTipText(MouseEvent e) {
		//    	System.out.println("In getToolTipText() "+props.isAnnotationPopupEnabled());
		int row = findRow(e.getX(), e.getY());

		if (row >= 0) {
			SequenceComponent sc = (SequenceComponent) getComponent(row);
			Point p = sc.getLocation();
			int col = sc.findColumn(e.getX() - p.x, e.getY() - p.y);

			if (col < 0) {
				return null;
			}

			Sequence seq = alignment.getSequence(row);
			Sequence.ColumnAnnotation ca = seq.getColumnAnnotation(col);
			//            if(ca == null)
			//            	System.out.println("col annotation null");
			//TODO changed the condition col >= 1 to col >= 0; In the original code, >=1 could be due to the index starting from 1
			if ((col >= 0) && (ca != null) && displayProperties.isAnnotationPopupEnabled()) {
				StringBuffer activityInfoStringBuffer;

				activityInfoStringBuffer = new StringBuffer("");
				activityInfoStringBuffer.append("<html> <font face = 'Courier' size = 5>");
				ca.getFormattedAnnotationInfo(seq, activityInfoStringBuffer);
				activityInfoStringBuffer.append("</font> </html>");

				return activityInfoStringBuffer.toString();
			} else {
				return null;
			}
		} else {
			return null;
		}
	}

	// MouseListener, MouseMotionListener interface
	private boolean isMousePressed = false;
	private int mouseRowIndex;
	private int mouseColIndex;

	@SuppressWarnings("deprecation")
	public void mousePressed(MouseEvent e) {

		mouseRowIndex = findRow(e.getX(), e.getY());

		if (mouseRowIndex >= 0) {
			SequenceComponent sequenceComponent = (SequenceComponent) getComponent(mouseRowIndex);
			Point p = sequenceComponent.getLocation();

			mouseColIndex = sequenceComponent.findColumn(e.getX() - p.x, e.getY() - p.y);
		} else {
			mouseColIndex = -1;
		}

		//TODO changed the condition for ColIndex to >=0 instead of >= 1
		isMousePressed = (mouseRowIndex >= 0) && (mouseColIndex >= 0);

		if ((e.getClickCount() == 1) && isMousePressed) {
			//			System.out.println("In MousePressed and ClickCount = 1 \t "+mouseRowIndex+" @ "+mouseColIndex);
			displayProperties.setSeqAnnEditing(false);
			displayProperties.setCursorHidden(false);
			alignment.deselectAllRulerSelections();
			alignment.deselectAllColumns(displayProperties);

			displayProperties.updateCursor(mouseRowIndex, mouseColIndex, e);
		}

		if ((e.getClickCount() >= 2) && isMousePressed) {
			Sequence seq = alignment.getSequence(mouseRowIndex);
			EditActivityAnnDialog editActivityAnnDialog = new EditActivityAnnDialog(owner, displayProperties,
					alignment, seq, mouseColIndex);

			editActivityAnnDialog.show();
		}
	}

	public void mouseReleased(MouseEvent arg0) {
		isMousePressed = false;
	}

	public void mouseDragged(MouseEvent arg0) {
		// TODO Need to add a lot; check the original; Currently just printing to proceed with the rest of the UI
		//		System.out.println("In AlignmentPanel: MouseDragged");
	}

	public void mouseClicked(MouseEvent arg0) {
		// Nothing to be done here
	}

	public void mouseEntered(MouseEvent arg0) {
		// Nothing to be done here
	}

	public void mouseExited(MouseEvent arg0) {
		// Nothing to be done here
	}

	public void mouseMoved(MouseEvent arg0) {
		// Nothing to be done here
	}

	public void mouseWheelMoved(MouseWheelEvent e) {
		// TODO Need to implement the incrementalVerticalScroll in
		// AlignmentFrame; Since there are multiple alignments, first
		// identifying the active tab and then moving the scroll is required
		// Currently only print sysout

		//		System.out.println("In AlignmentPanel: MouseWheelMoved");
		// owner.incrementalVerticalScroll(e.getWheelRotation());
	}

	// KeyListener interface implementation
	//	@Override
	public void keyPressed(KeyEvent arg0) {
		// TODO Auto-generated method stub

	}

	//	@Override
	public void keyReleased(KeyEvent arg0) {
		// TODO Auto-generated method stub

	}

	//	@Override
	public void keyTyped(KeyEvent arg0) {
		// TODO Auto-generated method stub

	}

	// AlignmentListener interface implementation
	public void alignmentSeqDeleted(Alignment alignment, int i, Sequence sequence) {
		if (alignment != this.alignment) {
			throw new RuntimeException("Bound to incorrect alignment");
		}
		remove(i);
		revalidate();
		repaint();

	}

	public void alignmentSeqInserted(Alignment alignment, int i) {
		if (this.alignment != alignment) {
			throw new RuntimeException("Bound to incorrect alignment");
		}

		JComponent sequenceComponent = new SequenceComponent(owner, alignment.getSequence(i), displayProperties);
		sequenceComponent.setLayout(new BoxLayout(sequenceComponent, BoxLayout.Y_AXIS));
		add(sequenceComponent, i);
		revalidate();
		repaint();
	}

	public void alignmentSeqSwapped(Alignment alignment, int i, int j) {
		if (alignment != this.alignment) {
			throw new RuntimeException("bound to incorrect alignment");
		}
		((SequenceComponent) getComponent(i)).setSequence(alignment.getSequence(i));
		((SequenceComponent) getComponent(j)).setSequence(alignment.getSequence(j));
	}

	public void alignmentNameChanged(Alignment alignment) {
		// Nothing to be done here
	}

	public void alignmentSeqActivityChanged(Alignment alignment, Sequence sequence) {
		// Nothing to be done here
	}

	// Scrollable interface implementation

	public Dimension getPreferredScrollableViewportSize() {
		return getPreferredSize();
	}

	public int getScrollableBlockIncrement(Rectangle visibleRectange, int orientation, int direction) {
		return orientation == SwingConstants.VERTICAL ? visibleRectange.height : visibleRectange.width;
	}

	public boolean getScrollableTracksViewportHeight() {
		return false;
	}

	public boolean getScrollableTracksViewportWidth() {
		return false;
	}

	public int getScrollableUnitIncrement(Rectangle visibleRect, int orientation, int direction) {
		return orientation == SwingConstants.VERTICAL ? displayProperties.getActivityHeight() : displayProperties
				.getActivityWidth();
	}

}
