package org.processmining.plugins.tracealignment.ui;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Rectangle;
import java.util.HashSet;
import java.util.Set;

import javax.swing.JPanel;

@SuppressWarnings("serial")
public class ColumnAnnotationComponent extends JPanel implements AlignmentListener, SequenceListener,
		DisplayPropertiesListener {

	private final Alignment alignment;
	private final DisplayProperties displayProperties;

	// active columns
	private Set<AnnotationDetails>[] columnMapSetArray = null;

	// maximum number of symbols per column
	private int maxNoSymbolsInColumn = 0;

	// size
	private int size = 0;

	public ColumnAnnotationComponent(Alignment alignment, DisplayProperties displayProperties) {
		super();

		this.alignment = alignment;
		this.displayProperties = displayProperties;

		this.alignment.addListener(this);
		this.displayProperties.addListener(this);

		refreshSize(false);
		refreshColumns(false);

		int noSequencesInAlignment = alignment.getNoSequences();
		for (int i = noSequencesInAlignment - 1; i >= 0; i--) {
			alignment.getSequence(i).addListener(this);
		}
		setBackground(Color.white);
	}

	private void refreshSize(boolean redraw) {
		int newSize = alignment.maxLength();

		if (size != newSize) {
			size = newSize;
			if (redraw) {
				revalidate();
				repaint();
			}
		}
	}

	@SuppressWarnings("unchecked")
	private void refreshColumns(boolean redraw) {
		// build a list of columns with annotation

		Set<AnnotationDetails>[] newColumnMapSetArray = new Set[size];
		int newMaxNoSymbolsInColumn = 0;

		int noSequencesInAlignment = alignment.getNoSequences();
		for (int i = noSequencesInAlignment - 1; i >= 0; i--) {
			Sequence seq = alignment.getSequence(i);

			if (seq.getColumnAnnotationsCount() > 0) {
				Sequence.ColumnAnnotation[] ca = seq.getColumnAnnotations();

				for (int k = ca.length - 1; k >= 0; k--) {
					int col = ca[k].getColumn();
					if ((col < 0) || (col >= size)) {
						System.err.println("col:" + col + " is out of bounds for col annotation " + k);
					}
					if (newColumnMapSetArray[col] == null) {
						newColumnMapSetArray[col] = new HashSet<AnnotationDetails>(3);
					}
					newColumnMapSetArray[col].add(new AnnotationDetails(ca[k].getSymbol(), ca[k].getColor()));
					newMaxNoSymbolsInColumn = Math.max(newMaxNoSymbolsInColumn, newColumnMapSetArray[col].size());
				}
			}
		}
		// figure out if anything has changed

		boolean changed = false;

		if ((columnMapSetArray == null) || (maxNoSymbolsInColumn != newMaxNoSymbolsInColumn)
				|| (columnMapSetArray.length != newColumnMapSetArray.length)) {
			changed = true;
		} else {
			for (int i = columnMapSetArray.length - 1; i >= 0; i--) {
				if ((columnMapSetArray[i] == null) || (newColumnMapSetArray[i] == null)) {
					if (columnMapSetArray[i] != newColumnMapSetArray[i]) {
						changed = true;
						break;
					}
				} else if (!columnMapSetArray[i].equals(newColumnMapSetArray[i])) {
					changed = true;
					break;
				}
			}
		}
		// set the map if things have changed

		if (changed) {
			columnMapSetArray = newColumnMapSetArray;
			maxNoSymbolsInColumn = newMaxNoSymbolsInColumn;
			if (redraw) {
				revalidate();
				repaint();
			}
		}
	}

	public Dimension getPreferredSize() {
		int height = displayProperties.getActivityHeight() * maxNoSymbolsInColumn;
		int width;

		//TODO changed from original
		width = displayProperties.getActivityWidth() * size;

		return new Dimension(width, height);
	}

	public void paintComponent(Graphics g) {
		if (!displayProperties.showActivityAnnPanel()) {
			return;
		}
		super.paintComponent(g);
		if (maxNoSymbolsInColumn < 1) {
			return;
		}

		Rectangle clip = g.getClipBounds();

		int activityHeight = displayProperties.getActivityHeight();
		int activityWidth = displayProperties.getActivityWidth();
		//        int spacingWidth = displayProperties.getSpacingWidth();

		int xFontOffset = displayProperties.getFontXOffset();
		int yFontOffset = displayProperties.getFontYOffset();
		int blockHeight = activityHeight * maxNoSymbolsInColumn;

		int start;
		int end;

		//TODO changed from original basically to do with isMapping
		start = clip.x / activityWidth;
		end = (clip.x + clip.width) / activityWidth;

		if (end >= columnMapSetArray.length) {
			end = columnMapSetArray.length - 1;
		}
		g.setFont(displayProperties.getFont());

		//		int spacingAdjustment = 0;
		//        int previousColumn = -1;

		//TODO changed from original quite a lot
		for (int index = start; index <= end; index++) {
			if (columnMapSetArray[index] != null) {
				int x = index * activityWidth;

				if (clip.intersects(x, 0, activityWidth, blockHeight)) {
					int j = 0;

					for (AnnotationDetails ad : columnMapSetArray[index]) {
						int y = j * activityHeight;

						g.setColor(ad.color);
						g.fillOval(x, y, activityWidth - 1, activityHeight - 1);
						g.setColor(displayProperties.inverseRGB(ad.color));
						g.drawString(ad.symbol, x + xFontOffset, y + yFontOffset);
						j++;
					}
				}
			}
		}
	}

	private static class AnnotationDetails {
		public String symbol;
		public Color color;

		public AnnotationDetails(String symbol, Color color) {
			this.symbol = symbol;
			this.color = color;
		}

		public boolean equals(Object other) {
			if (other instanceof AnnotationDetails) {
				AnnotationDetails o = (AnnotationDetails) other;

				return symbol.equals(o.symbol) && color.equals(o.color);
			}
			return false;
		}

		public int hashCode() {
			return symbol.hashCode() + 11 * color.hashCode();
		}
	}

	//AlignmentListener interface implementation
	public void alignmentSeqDeleted(Alignment alignment, int i, Sequence sequence) {
		if (alignment != this.alignment) {
			throw new RuntimeException("bound to incorrect alignment");
		}
		sequence.removeListener(this);
		refreshSize(true);
		refreshColumns(true);
	}

	public void alignmentSeqInserted(Alignment alignment, int i) {
		if (alignment != this.alignment) {
			throw new RuntimeException("Bound to incorrect alignment");
		}
		alignment.getSequence(i).addListener(this);
		refreshSize(true);
		refreshColumns(true);
	}

	public void alignmentNameChanged(Alignment alignment) {
		// Nothing to be done here
	}

	public void alignmentSeqActivityChanged(Alignment alignment, Sequence sequence) {
		// Nothing to be done here
	}

	public void alignmentSeqSwapped(Alignment alignment, int i, int j) {
		// Nothing to be done here
	}

	// SequenceListener interface implementation
	public void sequenceActivityChanged(Sequence aaseq) {
		refreshSize(true);
		refreshColumns(true);
	}

	public void sequenceColumnAnnotationsChanged(Sequence sequence, int column) {
		refreshColumns(true);
	}

	public void sequenceAnnotationChanged(Sequence sequence) {
		// Nothing to be done here
	}

	public void sequenceColorChanged(Sequence sequence) {
		// Nothing to be done here
	}

	public void sequenceLineAnnotationsChanged(Sequence aaseq) {
		// Nothing to be done here
	}

	public void sequenceNameChanged(Sequence sequence, String oldName) throws Exception {
		// Nothing to be done here
	}

	// DisplayPropertiesListener interface implementation

	public void displayFontChanged(DisplayProperties dp) {
		if (dp != displayProperties) {
			throw new RuntimeException("Bound to incorrect DisplayProperties");
		}
		revalidate();
		repaint();
	}

	public void displayAnnViewChanged(DisplayProperties dp, Sequence sequence, boolean show) {
		// Nothing to be done here
	}

	public void displayColorSchemeChanged(DisplayProperties dp, ColorScheme old) {
		// Nothing to be done here
	}

	public void displayGroupEditingChanged(DisplayProperties dp) {
		// Nothing to be done here
	}

	public void displayHighlightsChanged(DisplayProperties dp, Sequence sequence) {
		// Nothing to be done here
	}

	public void displayHighlightsChanged(DisplayProperties dp, Sequence[] sequenceArray) {
		// Nothing to be done here
	}

	public void displayOverwriteChanged(DisplayProperties dp) {
		// Nothing to be done here
	}

	public void displayRenderGapsChanged(DisplayProperties dp) {
		// Nothing to be done here
	}

	public void activityBackgroundChanged(DisplayProperties dp) {
		// Nothing to be done here
	}

	public void displaySeqSelectChanged(DisplayProperties dp, Sequence sequence, boolean select) {
		// Nothing to be done here
	}

}
