package org.processmining.plugins.tracealignment.ui;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Rectangle;
import java.awt.event.MouseEvent;
import java.text.DecimalFormat;

import javax.swing.JPanel;

@SuppressWarnings("serial")
public class ConsensusComponent extends JPanel implements SequenceListener, DisplayPropertiesListener {

	// underlying sequence
	private ConsensusSequence consensus = null;

	// rendering properties
	private final DisplayProperties displayProperties;

	public ConsensusComponent(ConsensusSequence consensus, DisplayProperties displayProperties) {
		super();
		setBackground(Color.white);
		this.displayProperties = displayProperties;
		displayProperties.addListener(this);
		setConsensus(consensus);

		setToolTipText("");
	}

	public void setConsensus(ConsensusSequence consensus) {
		if (this.consensus != null) {
			this.consensus.removeListener(this);
		}
		this.consensus = consensus;
		revalidate();
		repaint();
	}

	public Dimension getPreferredSize() {
		// Increased multiplier from 3 to 4 to leave some headroom
		int height = displayProperties.getActivityHeight() * 4;

		//		int width= getLayout().preferredLayoutSize(this).width;
		int width = displayProperties.getActivityWidth() * consensus.getLengthWithGaps();
		// System.err.println("Consensus preferred size:"+width+" "+height);
		return new Dimension(width, height);
	}

	public Color getConsensusColorAt(int i) {
		//		Color consensusColor = Color.LIGHT_GRAY;
		return Color.LIGHT_GRAY;
	}

	public String getToolTipText(MouseEvent e) {
		if (!displayProperties.isAnnotationPopupEnabled()) {
			return null;
		}

		// find out which col we are talking...
		int x = e.getX();
		int total = displayProperties.getMaxLength();
		int col = x / displayProperties.getActivityWidth();

		// out of range means nein!
		if ((col < 0) || (col >= total)) {
			return null;
		}

		//TODO changed 
		double value = getConservationScoreAtColumnIndex(col);
		DecimalFormat decimalFormat = new DecimalFormat("#.##");
		return decimalFormat.format(value);
	}

	public double getConservationScoreAtColumnIndex(int i) {
		return consensus.getInformationInColumnGapsIncluded()[i];
	}

	public void paint(Graphics g) {
		g.setColor(Color.white);

		Dimension d = getPreferredSize(); // need to reset size just in case
		// Dimension d = getSize();
		this.setSize(d); // alignment length changed

		g.fillRect(0, 0, d.width, d.height);

		int len = consensus.getLengthWithGaps();

		// System.out.print("IN paint length:"+len);
		int activityWidth = displayProperties.getActivityWidth();
		int activityHeight = displayProperties.getActivityHeight();
		int blockHeight = 3 * activityHeight;
		boolean render_gaps = displayProperties.isGapRendered();

		g.setFont(displayProperties.getFont());
		int start;
		int end;

		Rectangle clip = g.getClipBounds();
		if (clip == null) {
			//			ErrorDialog.showErrorDialog(this, "CLip is Null");
			start = 0;
			end = len - 1;

			clip = new Rectangle(0, 0, getPreferredSize().width, getPreferredSize().height);
		} else {
			start = (clip.x / activityWidth); // skip dummy residue
			end = (clip.x + clip.width) / activityWidth;
		}

		if (end >= len) {
			end = len - 1;
		}

		if (start < 0) {
			start = 0; // skip dummy residue
		}

		// ////////////////////////////////////////
		// Draw a red line at the beginning of component
		// this is useful for debugging incorrect placement
		// of the component
		g.setColor(new Color(156, 1, 1));
		g.drawLine(0, activityHeight, d.width, activityHeight);
		// ///////////////////////////////////////////////////

		// iterate over current clip and paint
		for (int index = start; index <= end; index++) {
			int x = (index) * activityWidth;
			if (clip.intersects(x, 0, activityWidth, blockHeight)) {
				String encodedActivity = consensus.getEncodedActivity(index);
				double consensusValue = getConservationScoreAtColumnIndex(index);
				Color consensusColor = getConsensusColorAt(index);
				int height = (int) (2 * activityHeight * Math.abs(consensusValue));

				g.setColor(consensusColor);
				g.fillRect(x, 3 * activityHeight - height, activityWidth, height);
				g.setColor(Color.GRAY);
				g.drawRect(x, 3 * activityHeight - height, activityWidth, height);

				if ((!encodedActivity.contains("-") || render_gaps)) {
					g.setColor(Color.black);
					g.drawString(encodedActivity, x + displayProperties.getFontXOffset(), 3 * activityHeight
							+ displayProperties.getFontYOffset());
				}
			}
		}
	}

	// SequenceListener interface implementation
	public void sequenceActivityChanged(Sequence aaseq) {
		if (aaseq != consensus) {
			throw new RuntimeException("bound to incorrect Sequence");
		}
		revalidate();
		repaint();
	}

	public void sequenceAnnotationChanged(Sequence sequence) {
		// Nothing to be done here
	}

	public void sequenceColorChanged(Sequence sequence) {
		// Nothing to be done here
	}

	public void sequenceColumnAnnotationsChanged(Sequence sequence, int column) {
		// Nothing to be done here
	}

	public void sequenceLineAnnotationsChanged(Sequence aaseq) {
		// Nothing to be done here
	}

	public void sequenceNameChanged(Sequence sequence, String oldName) throws Exception {
		// Nothing to be done here
	}

	// DisplayPropertiesListener interface implementation
	public void displayAnnViewChanged(DisplayProperties dp, Sequence sequence, boolean show) {
		// Nothing to be done here
	}

	public void displayColorSchemeChanged(DisplayProperties dp, ColorScheme old) {
		// Nothing to be done here
	}

	public void displayFontChanged(DisplayProperties dp) {
		if (dp != displayProperties) {
			throw new RuntimeException("bound to incorrect DisplayProperties");
		}
		revalidate();
		repaint();
	}

	public void displayGroupEditingChanged(DisplayProperties dp) {
		// Nothing to be done here
	}

	public void displayHighlightsChanged(DisplayProperties dp, Sequence sequence) {
		// Nothing to be done here
	}

	public void displayHighlightsChanged(DisplayProperties dp, Sequence[] sequenceArray) {
		// Nothing to be done here
	}

	public void displayOverwriteChanged(DisplayProperties dp) {
		// Nothing to be done here
	}

	public void displayRenderGapsChanged(DisplayProperties dp) {
		if (dp != displayProperties) {
			throw new RuntimeException("bound to incorrect DisplayProperties");
		}
		revalidate();
		repaint();
	}

	public void activityBackgroundChanged(DisplayProperties dp) {
		// Nothing to be done here
	}

	public void displaySeqSelectChanged(DisplayProperties dp, Sequence sequence, boolean select) {
		// TODO Auto-generated method stub

	}

}
