package org.processmining.plugins.tracealignment.ui;

import java.awt.Color;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.event.MouseEvent;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import org.processmining.plugins.tracealignment.ui.Sequence.LineAnnotation;

@SuppressWarnings("unused")
public class DisplayProperties implements AlignmentListener, SequenceListener {

	// the encodingLength of activities
	private final int encodingLength;

	private final Alignment alignment;

	// Activity text font
	private Font font;

	// Activity font metrics
	private FontMetrics fontMetrics;

	// Activity width, height
	private int activityWidth, activityHeight;

	// Activity font x-offset y-offset
	private int fontXOffset, fontYOffset;

	// spacing 
	private int spacingWidth;

	// cursor position details
	int cursorRow = 0, cursorColumn = 0;
	int previousCursorRow = 0, previousCursorColumn = 0;

	// Set of selected and hidden sequences

	Set<Sequence> selectedSequencesSet = new HashSet<Sequence>();
	Set<Sequence> hiddenSequencesSet = new HashSet<Sequence>();

	// Set of sequences with line annotations
	Set<Sequence> sequenceAnnotionSet = new HashSet<Sequence>();
	private final Set<LineAnnotation> hiddenLineAnnotationSet = new HashSet<LineAnnotation>();

	private boolean alignmentRecentlyEdited = false;

	// is cursor in RulerAnnotationComponent or SequenceAnnComponent?
	private boolean cursorHidden = false;

	// Set whether fast rendering mode or not
	private boolean fastRender = true;

	// Set whether to render gaps or not
	private boolean renderGaps = true;

	// Set whether encodedActivity is to be displayed or not
	private boolean renderEncodedActivity = true;

	// Set whether the background for activity is square or not
	private boolean isActivityBackGroundSquare = true;

	private final boolean showActivityAnnPanel = true;

	// Set whether group editing
	private final boolean isGroupEditing = false;
	private boolean isRulerEditing = false;

	private final boolean isMapping = false; // if true, we are column mapping...always true when printing and when columns are hidden

	private final boolean annotationPopupEnabled = true;
	private boolean isSeqAnnEditing = false;

	// Capture the sequences and the activity positions that are highlighted
	Map<Sequence, TreeSet<Integer>> sequenceActivityHighLightsMap = new HashMap<Sequence, TreeSet<Integer>>();

	// List of listeners
	List<DisplayPropertiesListener> listenersList = new ArrayList<DisplayPropertiesListener>();

	public DisplayProperties(int encodingLength, Alignment alignment, Font font, FontMetrics fontMetrics,
			boolean fastRender, boolean renderGaps) {
		this.encodingLength = encodingLength;
		this.alignment = alignment;
		this.fastRender = fastRender;
		this.renderGaps = renderGaps;

		setFont(font, fontMetrics);

		alignment.addListener(this);

		int noSequencesInAlignment = alignment.getNoSequences();
		for (int i = 0; i < noSequencesInAlignment; i++) {
			alignment.getSequence(i).addListener(this);
		}
	}

	private void setFont(Font font, FontMetrics fontMetrics) {
		this.font = font;
		this.fontMetrics = fontMetrics;

		if (fastRender) {
			activityWidth = fontMetrics.stringWidth("A") * encodingLength + 3;
			fontXOffset = 3;
		} else {
			activityWidth = fontMetrics.stringWidth("A") * encodingLength + 3;
			fontXOffset = 1;
		}

		spacingWidth = (int) (activityWidth * 1.5);
		activityHeight = fontMetrics.getMaxAscent() + fontMetrics.getMaxDescent() + 5;
		fontYOffset = fontMetrics.getMaxAscent() + 3;

		for (DisplayPropertiesListener listener : listenersList) {
			listener.displayFontChanged(this);
		}
	}

	public Color inverseRGB(Color c) {
		int brightness = c.getRed() + c.getGreen() + c.getBlue();

		if (brightness < 382) {
			return Color.white;
		} else {
			return Color.black;
		}
	}

	public boolean isAnnotationPopupEnabled() {
		return annotationPopupEnabled;
	}

	public boolean getSeqSelect(Sequence s) {
		return selectedSequencesSet.contains(s);
	}

	public boolean isMapping() {
		return isMapping;
	}

	public void setSeqAnnEditing(boolean b) {
		isSeqAnnEditing = b;
	}

	public Sequence[] getAllSelected() {
		return getAllSelected(false);
	}

	public Sequence[] getAllSelected(boolean inViewOrder) {
		Sequence[] seqs = new Sequence[selectedSequencesSet.size()];
		int cnt = 0;
		if (inViewOrder) {
			int total = alignment.getNoSequences();
			for (int i = 0; i < total; i++) {
				Sequence s = alignment.getSequence(i);
				if (selectedSequencesSet.contains(s)) {
					seqs[cnt++] = s;
				}
			}
		} else {
			for (Sequence s : selectedSequencesSet) {
				seqs[cnt++] = s;
			}
		}
		return seqs;
	}

	public int getSelectedCount() {
		return selectedSequencesSet.size();
	}

	// select a particular sequence
	public void setSeqSelect(Sequence s, boolean select) {
		boolean notify = false;

		if (select && selectedSequencesSet.add(s)) {
			notify = true;
		} else if (!select && selectedSequencesSet.remove(s)) {
			notify = true;
		}
		if (notify) {
			for (DisplayPropertiesListener listener : listenersList) {
				listener.displaySeqSelectChanged(this, s, select);
			}
		}
	}

	/**
	 * Deselect all sequences
	 */

	public void clearSelections() {
		if (selectedSequencesSet.size() < 1) {
			return;
		}
		ArrayList<Sequence> selectedSequencesList = new ArrayList<Sequence>(selectedSequencesSet.size());

		selectedSequencesList.addAll(selectedSequencesSet);
		selectedSequencesSet.clear();
		for (Sequence s : selectedSequencesList) {
			for (DisplayPropertiesListener listener : listenersList) {
				listener.displaySeqSelectChanged(this, s, false);
			}
		}
	}

	// show a line annotations associated with a particular sequence

	public void setAnnView(Sequence s, boolean show) {
		boolean notify = true;

		if (show && sequenceAnnotionSet.add(s)) {
			notify = true;
		} else if (!show && sequenceAnnotionSet.remove(s)) {
			notify = true;
		}
		if (notify) {
			for (DisplayPropertiesListener listener : listenersList) {
				listener.displayAnnViewChanged(this, s, show);
			}
		}
	}

	public Set<LineAnnotation> getHiddenLineAnnotations() {
		return hiddenLineAnnotationSet;
	}

	/**
	 * Get a list of the line annotations that are visible for the given
	 * sequence.
	 * 
	 * @param sequence
	 * @return
	 */
	public List<LineAnnotation> getVisibleLineAnnotations(Sequence sequence) {
		List<LineAnnotation> viewableLineAnnotationList = new ArrayList<LineAnnotation>(Arrays.asList(sequence
				.getLineAnnotations()));
		viewableLineAnnotationList.removeAll(this.getHiddenLineAnnotations());
		return viewableLineAnnotationList;
	}

	/**
	 * Get a list of the line annotations that are hidden for the given sequence
	 * 
	 * @param sequence
	 * @return
	 */
	public List<LineAnnotation> getHiddenLineAnnotations(Sequence sequence) {
		List<LineAnnotation> sequenceHiddenLineAnnotationList = new ArrayList<LineAnnotation>();

		for (LineAnnotation lineAnnotation : hiddenLineAnnotationSet) {
			if (lineAnnotation.getSequence().equals(sequence)) {
				sequenceHiddenLineAnnotationList.add(lineAnnotation);
			}
		}

		return sequenceHiddenLineAnnotationList;
	}

	public void setCursorHidden(boolean b) {
		cursorHidden = b;
	}

	public void updateCursor(int newCursorRow, int newCursorColumn) {
		updateCursor(newCursorRow, newCursorColumn, false, false);
	}

	public void updateCursor(int newCursorRow, int newCursorColumn, MouseEvent e) {
		// is meta down
		updateCursor(newCursorRow, newCursorColumn, e.isShiftDown(), e.isControlDown() || e.isMetaDown());
	}

	public void updateCursor(int newCursorRow, int newCursorColumn, boolean isShiftDown, boolean isControlDown) {
		//		System.out.println("updating cursor: isShiftDown="+isShiftDown+" @ isControlDown="+isControlDown);
		previousCursorRow = cursorRow;
		previousCursorColumn = cursorColumn;
		cursorRow = newCursorRow;
		cursorColumn = newCursorColumn;

		if (!cursorHidden) {
			if (!isShiftDown && !isControlDown) {
				clearHighlights();
			}

			int activityIndex = newCursorColumn;

			if (!isShiftDown && !isControlDown) {
				setSeqHighlight(alignment.getSequence(cursorRow), cursorColumn, true);
				alignment.setLastActivityIndexSelected(cursorRow, activityIndex);
			} else if (isShiftDown && !isControlDown) {
				int lastActivityIndex = alignment.getLastActivityIndexSelected();
				int lastActivityRow = alignment.getLastActivityRowSelected();
				if ((lastActivityIndex != -1) && (lastActivityRow == cursorRow)) {
					int sgn = sign(lastActivityIndex - activityIndex);
					for (int i = activityIndex; i != (lastActivityIndex + sgn); i = i + sgn) {
						setSeqHighlight(alignment.getSequence(cursorRow), i, true);
					}
					alignment.setLastActivityIndexSelected(cursorRow, activityIndex);
				} else {
					setSeqHighlight(alignment.getSequence(cursorRow), cursorColumn, true);
					alignment.setLastActivityIndexSelected(cursorRow, activityIndex);
				}
			} else if (!isShiftDown && isControlDown) {
				if (isSeqHighlight(alignment.getSequence(cursorRow), activityIndex)) {
					setSeqHighlight(alignment.getSequence(cursorRow), cursorColumn, false);
				} else {
					setSeqHighlight(alignment.getSequence(cursorRow), cursorColumn, true);
					alignment.setLastActivityIndexSelected(cursorRow, activityIndex);
				}
			} else if (isShiftDown && isControlDown) {
				setSeqHighlight(alignment.getSequence(cursorRow), cursorColumn, true);
				alignment.setLastActivityIndexSelected(cursorRow, activityIndex);
			}

			alignment.clearLastRulerAnnIndexSelected();
			alignment.clearLastColumnIndexSelected();
		}
	}

	public int sign(int a) {
		if (a == 0) {
			return 1;
		} else {
			return a / Math.abs(a);
		}
	}

	public boolean isSeqHighlight(Sequence s, int idx) {
		TreeSet<Integer> highLightedIndicesSet = sequenceActivityHighLightsMap.get(s);
		if (highLightedIndicesSet == null) {
			return false;
		}

		return (highLightedIndicesSet.contains(new Integer(idx)));
	}

	public void setSeqHighlight(Sequence s, int idx, boolean highlight) {
		boolean notify = false;
		TreeSet<Integer> highLightedIndicesSet = sequenceActivityHighLightsMap.get(s);

		if (highlight) {
			if (highLightedIndicesSet == null) {
				highLightedIndicesSet = new TreeSet<Integer>();
				highLightedIndicesSet.add(new Integer(idx));
				sequenceActivityHighLightsMap.put(s, highLightedIndicesSet);
				notify = true;
			} else if (highLightedIndicesSet.add(new Integer(idx))) {
				notify = true;
			}
		} else {
			if ((highLightedIndicesSet != null) && highLightedIndicesSet.remove(new Integer(idx))) {
				if (highLightedIndicesSet.size() == 0) {
					sequenceActivityHighLightsMap.remove(s);
				}
				notify = true;
			}
		}
		if (notify) {
			for (DisplayPropertiesListener listener : listenersList) {
				listener.displayHighlightsChanged(this, s);
			}
		}
	}

	// highlight a particular sequence

	public void setSeqHighlight(Sequence s, int idx, boolean highlight, boolean notify) {
		boolean innerNotify = false;

		TreeSet<Integer> highLightedIndicesSet = sequenceActivityHighLightsMap.get(s);

		if (highlight) {
			if (highLightedIndicesSet == null) {
				highLightedIndicesSet = new TreeSet<Integer>();
				highLightedIndicesSet.add(new Integer(idx));
				sequenceActivityHighLightsMap.put(s, highLightedIndicesSet);
				innerNotify = true;
			} else if (highLightedIndicesSet.add(new Integer(idx))) {
				innerNotify = true;
			}
		} else {
			if ((highLightedIndicesSet != null) && highLightedIndicesSet.remove(new Integer(idx))) {
				if (highLightedIndicesSet.size() == 0) {
					sequenceActivityHighLightsMap.remove(s);
				}
				innerNotify = true;
			}
		}
		if (notify && innerNotify) {
			fireDisplayHighlightsChanged(s);
		}
	}

	public void fireDisplayHighlightsChanged(Sequence s) {
		for (DisplayPropertiesListener listener : listenersList) {
			listener.displayHighlightsChanged(this, s);
		}
	}

	public void fireDisplayHighlightsChanged(Sequence[] s) {
		for (DisplayPropertiesListener listener : listenersList) {
			listener.displayHighlightsChanged(this, s);
		}
	}

	public void setRenderGaps(boolean renderGaps) {
		if (this.renderGaps != renderGaps) {
			this.renderGaps = renderGaps;
			for (DisplayPropertiesListener listener : listenersList) {
				listener.displayRenderGapsChanged(this);
			}
		}
	}

	public void setRenderEncodedActivity(boolean showEncodedActivity) {
		if (renderEncodedActivity != showEncodedActivity) {
			renderEncodedActivity = showEncodedActivity;
			for (DisplayPropertiesListener listener : listenersList) {
				listener.displayRenderGapsChanged(this);
			}
		}
	}

	public void setActivityBackgroundSquare(boolean activityBackgroundSquare) {
		if (isActivityBackGroundSquare != activityBackgroundSquare) {
			isActivityBackGroundSquare = activityBackgroundSquare;
			for (DisplayPropertiesListener listener : listenersList) {
				listener.activityBackgroundChanged(this);
			}
		}
	}

	public boolean isEncodedActivityRendered() {
		return renderEncodedActivity;
	}

	public boolean isActivityBackgroundSquare() {
		return isActivityBackGroundSquare;
	}

	public int getMaxLength() {
		return alignment.maxLength();
	}

	public boolean getAnnView(Sequence s) {
		return sequenceAnnotionSet.contains(s);
	}

	public FontMetrics getFontMetrics() {
		return fontMetrics;
	}

	public int getCursorRow() {
		return cursorRow;
	}

	public int getPrevCursorRow() {
		return previousCursorRow;
	}

	public int getCursorColumn() {
		return cursorColumn;
	}

	public int getPrevCursorColumn() {
		return previousCursorColumn;
	}

	// fire sequence selected listeners
	public void fireSeqSelect(Sequence s) {
		boolean select = selectedSequencesSet.contains(s);
		for (DisplayPropertiesListener listener : listenersList) {
			listener.displaySeqSelectChanged(this, s, select);
		}
	}

	public void setVisibleLineAnnotation(Sequence sequence, LineAnnotation lineAnnotation) {
		hiddenLineAnnotationSet.remove(lineAnnotation);
		sequenceAnnotionSet.add(sequence);
		sequenceLineAnnotationsChanged(sequence);
	}

	public void setHiddenLineAnnotation(Sequence sequence, LineAnnotation lineAnnotation) {
		hiddenLineAnnotationSet.add(lineAnnotation);
		LineAnnotation[] lineAnnotationArray = sequence.getLineAnnotations();

		// Check whether all the lineAnnotations are hidden.  If all are hidden,
		// remove this sequence from seq_ann_view.
		boolean hasVisibleLineAnnotation = false;
		for (int i = 0; i < lineAnnotationArray.length; i++) {
			if (!hiddenLineAnnotationSet.contains(lineAnnotationArray[i])) {
				hasVisibleLineAnnotation = true;
				break;
			}
		}

		if (!hasVisibleLineAnnotation) {
			sequenceAnnotionSet.remove(sequence);
		}

		sequenceLineAnnotationsChanged(sequence);
	}

	/**
	 * Add a listener
	 * 
	 * @param listener
	 */
	public void addListener(DisplayPropertiesListener listener) {
		listenersList.add(listener);
	}

	/**
	 * Remove a listener
	 * 
	 * @param listener
	 */
	public void removeListener(DisplayPropertiesListener listener) {
		listenersList.remove(listener);
	}

	public boolean isGapRendered() {
		return renderGaps;
	}

	public boolean isGroupEditing() {
		return isGroupEditing;
	}

	public boolean isSeqAnnEditing() {
		return isSeqAnnEditing;
	}

	public void setAligmentRecentlyEdited(boolean b) {
		alignmentRecentlyEdited = b;
	}

	public void setSeqHidden(Sequence s, boolean hidden) {
		boolean notify = false;

		if (hidden && hiddenSequencesSet.add(s)) {
			notify = true;
		} else if (!hidden && hiddenSequencesSet.remove(s)) {
			notify = true;
		}
	}

	public boolean isSequenceHidden(int row) {
		return hiddenSequencesSet.contains(alignment.getSequence(row));
	}

	public boolean showActivityAnnPanel() {
		return showActivityAnnPanel;
	}

	public int getSpacingWidth() {
		return spacingWidth;
	}

	public Alignment getAlignment() {
		return alignment;
	}

	public void setRulerEditing(boolean b) {
		isRulerEditing = b;
	}

	public Font getFont() {
		return font;
	}

	public int getActivityWidth() {
		return activityWidth;
	}

	public int getActivityHeight() {
		return activityHeight;
	}

	public int getFontXOffset() {
		return fontXOffset;
	}

	public int getFontYOffset() {
		return fontYOffset;
	}

	public int getXForColumn(int column) {
		int w = getActivityWidth();
		//TODO haven't added the if condition on isMapping; check original
		return column * w;
	}

	/**
	 * Identify the column positions (activities) that are highlighted in the
	 * given sequence
	 * 
	 * @param sequence
	 * @return an array of column indices in ascending order
	 */
	public int[] getHighlights(Sequence sequence) {
		TreeSet<Integer> highlightedColumnIndicesSet = sequenceActivityHighLightsMap.get(sequence);

		if ((highlightedColumnIndicesSet == null) || (highlightedColumnIndicesSet.size() == 0)) {
			return null;
		}

		int[] highlightedColumnIndicesArray = new int[highlightedColumnIndicesSet.size()];

		int index = 0;
		for (Integer columnIndex : highlightedColumnIndicesSet) {
			highlightedColumnIndicesArray[index++] = columnIndex.intValue();
		}

		return highlightedColumnIndicesArray;
	}

	public void clearHighlights() {
		if (sequenceActivityHighLightsMap.size() < 1) {
			return;
		}
		ArrayList<Sequence> highlightedSequencesList = new ArrayList<Sequence>(sequenceActivityHighLightsMap.size());

		highlightedSequencesList.addAll(sequenceActivityHighLightsMap.keySet());
		sequenceActivityHighLightsMap.clear();
		for (Sequence s : highlightedSequencesList) {
			for (DisplayPropertiesListener listener : listenersList) {
				listener.displayHighlightsChanged(this, s);
			}
		}
	}

	// Alignment Listener implementation
	public void alignmentNameChanged(Alignment alignment) {
		// Nothing to be done for this here; Alignment frame is supposed to handle this
	}

	public void alignmentSeqDeleted(Alignment alignment, int i, Sequence sequence) {
		// TODO Auto-generated method stub
	}

	public void alignmentSeqInserted(Alignment alignment, int i) {
		// TODO Auto-generated method stub
	}

	public void alignmentSeqSwapped(Alignment alignment, int i, int j) {
		// Nothing to be done for this here;
	}

	public void alignmentSeqActivityChanged(Alignment alignment, Sequence sequence) {
		// Nothing to be done for this here;
	}

	// Sequence Listener interface implementation

	public void sequenceActivityChanged(Sequence sequence) {
		// Remove the sequence from highlight map
		sequenceActivityHighLightsMap.remove(sequence);

		// invoke the highlight changed method in the listener
		for (DisplayPropertiesListener listener : listenersList) {
			listener.displayHighlightsChanged(this, sequence);
		}
	}

	public void sequenceLineAnnotationsChanged(Sequence aaseq) {
		for (DisplayPropertiesListener listener : listenersList) {
			listener.displayAnnViewChanged(this, aaseq, true);
		}
	}

	public void sequenceNameChanged(Sequence sequence, String oldName) throws Exception {
		// Nothing to be done here
	}

	public void sequenceColorChanged(Sequence sequence) {
		// Nothing to be done here		
	}

	public void sequenceAnnotationChanged(Sequence sequence) {
		// Nothing to be done here
	}

	public void sequenceColumnAnnotationsChanged(Sequence sequence, int column) {
		// Nothing to be done here		
	}
}
