package org.processmining.plugins.tracealignment.ui;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;
import java.util.List;
import java.util.Random;

import javax.swing.JPanel;

import org.processmining.plugins.tracealignment.AlignmentFrame;
import org.processmining.plugins.tracealignment.ui.Sequence.LineAnnotation;

@SuppressWarnings( { "serial", "unused" })
public class SequenceComponent extends JPanel implements DisplayPropertiesListener, SequenceListener,
		ColorSchemeListener {
	
	private AlignmentFrame alignmentFrame;

	// Underlying sequence
	private Sequence sequence = null;

	// Rendering properties
	private final DisplayProperties displayProperties;

	// Rendering colors
	private Color[] foregroundColor, backgroundColor;

	// Whether the sequence is hidden or not
	public boolean collapsed = false;

	public SequenceComponent(AlignmentFrame alignmentFrame, Sequence sequence, DisplayProperties displayProperties) {
		super();
		this.alignmentFrame = alignmentFrame;
		this.sequence = sequence;
		this.displayProperties = displayProperties;

		displayProperties.addListener(this);
		//TODO
		// ColorScheme is not defined in DisplayProperties

		foregroundColor = backgroundColor = null;
		setSequence(sequence);
	}

	public void setSequence(Sequence sequence) {
		if (this.sequence != null) {
			this.sequence.removeListener(this);
		}

		this.sequence = sequence;
		this.sequence.addListener(this);

		foregroundColor = backgroundColor = null;
		revalidate();
		repaint();
	}

	public Sequence getSequence() {
		return sequence;
	}

	public Dimension getMinimumSize() {
		return getPreferredSize();
	}

	public Dimension getMaximumSize() {
		Dimension d = getPreferredSize();
		d.width = Integer.MAX_VALUE;
		return d;
	}

	public Dimension getPreferredSize() {
		int height = displayProperties.getActivityHeight();

		if (displayProperties.getAnnView(sequence)) {
			List<LineAnnotation> viewableLineAnnotations = displayProperties.getVisibleLineAnnotations(sequence);
			height *= 1 + viewableLineAnnotations.size();
		}

		int width;
		//TODO 
		// Check the original code for differences; Mostly dealing Mapping
		width = displayProperties.getActivityWidth() * sequence.getLengthWithGaps();

		if (collapsed) {
			height = 0;
		}

		return new Dimension(width, height);
	}

	public Dimension getSize() {
		Dimension d = super.getSize();
		if (collapsed) {
			d.height = 0;
		}
		return d;
	}

	/**
	 * Return the column number based on the coordinate location; Only the
	 * x-coordinate value is required to determine the column number
	 * 
	 * @param x
	 *            - the x-coordinate value
	 * @param y
	 *            - the y-coordinate value
	 * @return the column number or -1 in case of invalid value
	 */
	public int findColumn(int x, int y) {
		int index = x / displayProperties.getActivityWidth();

		index = index >= sequence.getLengthWithGaps() ? -1 : index;

		return index;
	}

	/**
	 * Get the coordinate location of an activity given the activity column
	 * (position) number
	 * 
	 * @param activityIndex
	 * @return Coordinate location
	 */
	public Point getActivityPosition(int activityIndex) {
		return new Point(activityIndex * displayProperties.getActivityWidth(), 0);
	}

	public DisplayProperties getDisplayProperties() {
		return displayProperties;
	}

	// caller must revalidate...
	public void setCollapsed(boolean collapsed) {
		this.collapsed = collapsed;
	}

	public boolean getCollapsed() {
		return collapsed;
	}

	private class RenderIterator {
		private final int activityWidth;
		private int currentIndex;
		private final int startIndex;
		private int endIndex;
		private int currentX;

		private final int[] highlights;
		private int hIndex;
		private boolean isThisActivityHighlighted = false;
		private final Color thisSequenceColor;

		private final Sequence.ColumnAnnotation[] columnAnnotationArray;
		private int columnAnnotationIndex;
		private Sequence.ColumnAnnotation thisActivityColumnAnnotation = null;

		private final boolean renderGaps;

		@SuppressWarnings("cast")
		public RenderIterator(Rectangle clip) {
			activityWidth = displayProperties.getActivityWidth();
			renderGaps = displayProperties.isGapRendered();

			int length = sequence.getLengthWithGaps();

			startIndex = 0;

			endIndex = (int) Math.floor(((float) clip.x + clip.width) / ((float) activityWidth)) + 1;

			if (endIndex >= length) {
				endIndex = length - 1;
			}

			thisSequenceColor = sequence.getColor();
			highlights = displayProperties.getHighlights(sequence);

			columnAnnotationArray = sequence.getColumnAnnotationsCount() > 0 ? sequence.getColumnAnnotations() : null;

			if (backgroundColor == null) {
				backgroundColor = new Color[length];
			}
			if (foregroundColor == null) {
				foregroundColor = new Color[length];
			}
		}

		private int currentIndex() {
			return currentIndex;
		}

		public boolean hasNext() {
			return currentIndex < endIndex;
		}

		public void reset() {
			currentIndex = startIndex - 1;
			//TODO
			currentX = (currentIndex) * activityWidth;
			if (highlights != null) {
				hIndex = 0;
			}
			if (columnAnnotationArray != null) {
				columnAnnotationIndex = 0;
			}
		}

		public int getIndex() {
			return currentIndex();
		}

		public int getX() {
			return currentX;
		}

		public int getStartIndex() {
			return startIndex;
		}

		public int getEndIndex() {
			return endIndex;
		}

		public int getStartX() {
			return (startIndex) * activityWidth;
		}

		public void next() {
			currentIndex++;
			currentX += activityWidth;

			if (highlights != null) {
				while ((hIndex < highlights.length) && (highlights[hIndex] < currentIndex)) {
					hIndex++;
				}
				isThisActivityHighlighted = (hIndex < highlights.length) && (highlights[hIndex] == currentIndex);
			}

			if (columnAnnotationArray != null) {
				while ((columnAnnotationIndex < columnAnnotationArray.length)
						&& (columnAnnotationArray[columnAnnotationIndex].getColumn() < currentIndex())) {
					columnAnnotationIndex++;
				}

				thisActivityColumnAnnotation = (columnAnnotationIndex < columnAnnotationArray.length)
						&& (columnAnnotationArray[columnAnnotationIndex].getColumn() == currentIndex()) ? columnAnnotationArray[columnAnnotationIndex]
						: null;
			}
		}

		public Sequence.ColumnAnnotation getColumnAnnotation() {
			return thisActivityColumnAnnotation;
		}

		public Color getBackgroundColor() {
			Color c;
			if (isThisActivityHighlighted) {
				if (thisSequenceColor != null) {
					c = thisSequenceColor;
				} else {
					c = foregroundColor[currentIndex()];
					if (c == null) {
						c = Color.black;
						foregroundColor[currentIndex()] = c;
					}
				}
			} else {
				c = backgroundColor[currentIndex()];
				if (c == null) {
					String encodedActivity = sequence.getEncodedActivity(currentIndex());
					c = alignmentFrame.getEncodedActivityColorMap().get(encodedActivity);
					if (encodedActivity.contains("-")) {
						c = Color.white;
					} else {
						if (c == null) {
							c = new Color(new Random().nextInt());
						}
					}
					backgroundColor[currentIndex()] = c;
				}
			}
			return c;
		}

		public Color getForegroundColor() {
			Color c;

			if (isThisActivityHighlighted) {
				c = backgroundColor[currentIndex()];
				if (c == null) {
					String encodedActivity = sequence.getEncodedActivity(currentIndex());
					c = alignmentFrame.getEncodedActivityColorMap().get(encodedActivity);
					if (encodedActivity.contains("-")) {
						c = Color.white;
					} else {
						if (c == null) {
							c = new Color(new Random().nextInt());
						}
					}
					backgroundColor[currentIndex()] = c;
				}
			} else {
				if (thisSequenceColor != null) {
					c = thisSequenceColor;
				} else {
					c = foregroundColor[currentIndex()];
					if (c == null) {
						c = Color.black;
						foregroundColor[currentIndex()] = Color.black;
					}
				}
			}
			return c;
		}

		private String getEncodedActivity() {
			String encodedActivity = sequence.getEncodedActivity(currentIndex());
			if (!encodedActivity.contains("-") || renderGaps) {
				return encodedActivity;
			}
			String blankSpace = " ";
			for (int i = 1; i <= alignmentFrame.getEncodingLength(); i++) {
				blankSpace += " ";
			}

			return blankSpace;
		}
	}

	public void paintComponent(Graphics g) {
		super.paintComponent(g);
		g.setColor(Color.white);

		Sequence.ColumnAnnotation ca = null;
		Color caColor = Color.white;

		Dimension d = getSize();
		g.fillRect(0, 0, d.width, d.height);

		Rectangle clip = g.getClipBounds();
		int activityWidth = displayProperties.getActivityWidth();
		int activityHeight = displayProperties.getActivityHeight();

		RenderIterator ri = new RenderIterator(clip);
		for (ri.reset(); ri.hasNext();) {
			ri.next();
			g.setColor(ri.getBackgroundColor());
			//adds background color to residue
			if (displayProperties.isActivityBackgroundSquare()) {
				g.fillRect(ri.getX(), 0, activityWidth, activityHeight);
			} else {
				//            	g.fillOval(ri.getX(), 0, activityWidth, activityHeight);
				g.fillRoundRect(ri.getX(), 0, activityWidth, activityHeight, activityWidth, activityHeight);
				//            g.fillOval(ri.getX(), 0, activityWidth, activityHeight);
			}
		}

		if (sequence.getColumnAnnotationsCount() > 0) {
			for (ri.reset(); ri.hasNext();) {
				ri.next();
				ca = ri.getColumnAnnotation();
				if ((ca != null) && ca.isShown()) {
					g.setColor(ca.getColor());
					g.fillOval(ri.getX(), 0, activityWidth - 1, activityHeight - 1);
				}
			}
		} else {
			ca = null;
		}

		//TODO draw found annotations

		// draw residue text

		int fontX = displayProperties.getFontXOffset();
		int fontY = displayProperties.getFontYOffset();

		g.setFont(displayProperties.getFont());

		for (ri.reset(); ri.hasNext();) {
			ri.next();
			String encodedActivity = ri.getEncodedActivity();

			if (!encodedActivity.contains(" ")) {
				ca = ri.getColumnAnnotation();
				if (ca != null) {
					caColor = ca.getColor();
				}
				Color color = (ca == null) || !ca.isShown() ? ri.getForegroundColor() : displayProperties
						.inverseRGB(caColor);

				g.setColor(color);
				if (displayProperties.isEncodedActivityRendered()) {
					g.drawString(encodedActivity, ri.getX() + fontX, fontY);
				}
			}
		}

		//TODO draw line annotations

		if (displayProperties.getAnnView(sequence)) {
			Sequence.LineAnnotation[] lineAnnotationArray = displayProperties.getVisibleLineAnnotations(sequence)
					.toArray(new Sequence.LineAnnotation[0]);

			g.setColor(Color.black);

			int y = 0;
			for (int k = 0; k < lineAnnotationArray.length; k++) {

				y += activityHeight;

				fontY = y + displayProperties.getFontYOffset();

				String ann = lineAnnotationArray[k].getAnnotation();

				if (ann.length() % alignmentFrame.getEncodingLength() != 0) {
					int remainder = ann.length() % alignmentFrame.getEncodingLength();
					for (int jj = 0; jj < remainder; jj++) {
						ann = ann.concat(" ");
					}
				}

				int max_idx = ann.length() / alignmentFrame.getEncodingLength();
				//				System.out.println("Max Idx: "+max_idx);

				for (ri.reset(); ri.hasNext();) {
					ri.next();
					int idx = ri.getIndex();
					//					System.out.println(idx);
					if (idx > max_idx) {
						break;
					}
					//TODO

					if ((idx + 1) < max_idx) {
						g.drawString(ann
								.substring(idx * alignmentFrame.getEncodingLength(), (idx + 1) * alignmentFrame.getEncodingLength()), ri
								.getX()
								+ fontX, fontY);
					} else {
						g.drawString(ann.substring(idx * alignmentFrame.getEncodingLength()), ri.getX() + fontX, fontY);
						//					g.drawString(ann.substring((idx - 1), idx),
						//							ri.getX() + fontX, fontY);
						//					g.drawString(ann,
						//							ri.getX() + fontX, fontY);
					}
				}
			}

		}

	}

	// DisplayPropertiesListener interface implementation
	public void displayAnnViewChanged(DisplayProperties dp, Sequence sequence, boolean show) {
//		System.out.println("in SequenceComponent:displayAnnViewChanged");
		if (dp != displayProperties) {
			throw new RuntimeException("Bound to incorrect DisplayProperties");
		}
		if (this.sequence == sequence) {
			revalidate();
			repaint();
		}
	}

	public void displayColorSchemeChanged(DisplayProperties dp, ColorScheme old) {
		// TODO Auto-generated method stub
	}

	public void displayHighlightsChanged(DisplayProperties dp, Sequence sequence) {
		if (dp != displayProperties) {
			throw new RuntimeException("Bound to incorrect DisplayProperties");
		}
		if (sequence == this.sequence) {
			revalidate();
			repaint();
		}
	}

	public void displayHighlightsChanged(DisplayProperties dp, Sequence[] sequenceArray) {
		if (dp != displayProperties) {
			throw new RuntimeException("Bound to incorrect DisplayProperties");
		}
		for (int i = 0; i < sequenceArray.length; i++) {
			if (sequence == sequenceArray[i]) {
				revalidate();
				repaint();
				return;
			}
		}
	}

	public void displayFontChanged(DisplayProperties dp) {
		if (dp != displayProperties) {
			throw new RuntimeException("Bound to incorrect DisplayProperties");
		}
		//TODO

		revalidate();
		repaint();
	}

	public void displayRenderGapsChanged(DisplayProperties dp) {
		if (dp != displayProperties) {
			throw new RuntimeException("Bound to incorrect DisplayProperties");
		}
		revalidate();
		repaint();
	}

	public void activityBackgroundChanged(DisplayProperties dp) {
		if (dp != displayProperties) {
			throw new RuntimeException("Bound to incorrect DisplayProperties");
		}
		revalidate();
		repaint();
	}

	public void displaySeqSelectChanged(DisplayProperties dp, Sequence sequence, boolean select) {
		// Nothing to be done here
	}

	public void displayOverwriteChanged(DisplayProperties dp) {
		// Nothing to be done here
	}

	public void displayGroupEditingChanged(DisplayProperties dp) {
		// Nothing to be done here
	}

	// SequenceListener interface implementation
	public void sequenceActivityChanged(Sequence sequence) {
		if (this.sequence != sequence) {
			throw new RuntimeException("Bound to incorrect Sequence");
		}
		foregroundColor = backgroundColor = null;
		revalidate();
		repaint();
	}

	public void sequenceColorChanged(Sequence sequence) {
		if (this.sequence != sequence) {
			throw new RuntimeException("Bound to incorrect Sequence");
		}
		foregroundColor = null;
		revalidate();
		repaint();
	}

	public void sequenceColumnAnnotationsChanged(Sequence sequence, int column) {
		if (this.sequence != sequence) {
			throw new RuntimeException("Bound to incorrect Sequence");
		}
		revalidate();
		repaint();

	}

	public void sequenceLineAnnotationsChanged(Sequence aaseq) {
		// TODO Auto-generated method stub

	}

	public void sequenceAnnotationChanged(Sequence sequence) {
		// Nothing to be done here
	}

	public void sequenceNameChanged(Sequence sequence, String oldName) throws Exception {
		// Nothing to be done here
	}

	// ColorSchemeListener interface implementation
	public void colorSchemeChanged(ColorScheme cs) {
		// TODO Auto-generated method stub

	}

}
