package org.processmining.plugins.tracealignment.ui;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;
import java.util.List;

import javax.swing.JPanel;

import org.processmining.plugins.tracealignment.AlignmentFrame;
import org.processmining.plugins.tracealignment.ui.Sequence.LineAnnotation;

@SuppressWarnings("serial")
public class SequenceNameComponent extends JPanel implements SequenceListener, DisplayPropertiesListener {

	public final static int Spacing = 5;

	/*
	 * Expandable line annotation box location in this component's space of the
	 * last drawn expand box (if any)
	 */
	private int expandBoxX, expandBoxY, expandBoxWidth, expandBoxHeight;

	// Underlying sequence
	private Sequence sequence = null;

	// Underlying sequence display properties
	private final DisplayProperties displayProperties;

	private Color sequenceColor = null;

	boolean collapsed = false;

	AlignmentFrame owner;

	public SequenceNameComponent(AlignmentFrame owner, Sequence sequence, DisplayProperties props) {
		super();
		setBackground(Color.white);
		this.owner = owner;
		displayProperties = props;
		props.addListener(this);
		setSequence(sequence);
		sequenceColor = sequence.getColor();
		if ((sequenceColor == null) || (sequenceColor == Color.white)) {
			sequenceColor = Color.black;
		}
	}

	public void setSequence(Sequence sequence) {
		if (this.sequence != null) {
			this.sequence.removeListener(this);
		}
		this.sequence = sequence;
		this.sequence.addListener(this);
		revalidate();
		repaint();
	}

	public void setCollapsed(boolean collapsed) {
		this.collapsed = collapsed;
	}

	public Sequence getSequence() {
		return sequence;
	}

	// default size
	public Dimension getMinimumSize() {
		return getPreferredSize();
	}

	public Dimension getPreferredSize() {
		FontMetrics fontMetrics = displayProperties.getFontMetrics();

		int height = displayProperties.getActivityHeight();

		if (displayProperties.getAnnView(sequence)) {
			List<LineAnnotation> viewableLineAnnotationsList = displayProperties.getVisibleLineAnnotations(sequence);
			height *= 1 + viewableLineAnnotationsList.size();
		}

		int width = 2 * displayProperties.getActivityWidth();
		//TODO; Changed from original 
		String displayName = sequence.getName();

		if (displayName != null) {
			width += fontMetrics.stringWidth(displayName) + 2;
		}

		width += 30;

		if (collapsed) {
			height = 0;
		}

		return new Dimension(width, height);
	}

	public Dimension getSize() {
		Dimension d = super.getSize();
		if (collapsed) {
			d.height = 0;
		}

		return d;
	}

	public Dimension getMaximumSize() {
		//System.out.println("Inside seq name comp - getMaxSize");
		Dimension d = getPreferredSize();
		//System.out.println("Received Dimension:"+d);
		d.width = Integer.MAX_VALUE;
		//System.out.println("Setting width in Dimension with int max value:"+d);
		return d;
	}

	public int findLineAnnotation(int x, int y) {
		if (!displayProperties.getAnnView(sequence)) {
			return -1;
		}

		int lineAnnotationCount = sequence.getLineAnnotationsCount();

		if (lineAnnotationCount < 1) {
			return -1;
		}
		return (int) Math.floor(((double) y) / ((double) displayProperties.getActivityHeight())) - 1;
	}

	protected Color getSeqColor() {
		//TODO changed from original
		return sequenceColor;
	}

	public boolean pointInExpandBox(Point p) {
		if (sequence.getLineAnnotationsCount() <= 0) {
			return false;
		}

		Rectangle r = new Rectangle(expandBoxX, expandBoxY, expandBoxWidth, expandBoxHeight);
		return r.contains(p);
	}

	public void paint(Graphics g) {
		boolean selected = displayProperties.getSeqSelect(sequence);
		g.setColor(selected ? getSeqColor() : Color.white);
		Dimension d = getSize();

		g.fillRect(0, 0, d.width, d.height);

		Font font = displayProperties.getFont();

		int x = Spacing + displayProperties.getActivityWidth() * 2;
		int y = 0;
		int fontY = y + displayProperties.getFontYOffset();
		int activityHeight = displayProperties.getActivityHeight();

		g.setFont(font);

		boolean showAnnotation = displayProperties.getAnnView(sequence);

		String str = sequence.getName();
		if (sequence.getNoDuplicates() > 0) {
			g.setColor(Color.lightGray);
			int width = str.length() - 1;
			if (width <= 1) {
				width++;
			}
			g.fill3DRect(x, y, width * font.getSize(), activityHeight, false);
		}

		g.setColor(selected ? Color.white : getSeqColor());
		g.drawString(str, x, fontY);

		//TODO removed one condition in the if clause displayProperties.isPrinting()
		if ((sequence.getLineAnnotationsCount() > 0)
				&& (displayProperties.getVisibleLineAnnotations(sequence).size() > 0)) {
			String aStr = (showAnnotation ? "-" : "+");
			x = Spacing + displayProperties.getActivityWidth() / 2;
			g.drawString(aStr, x, fontY);

			x = Spacing + 1;
			y = 5 + 1;
			FontMetrics fontMetrics = displayProperties.getFontMetrics();
			expandBoxX = x;
			expandBoxY = y;
			expandBoxWidth = fontMetrics.stringWidth(aStr) + 5;
			expandBoxHeight = fontMetrics.getAscent() - 1;
			g.drawRect(expandBoxX, expandBoxY, expandBoxWidth, expandBoxHeight);
		}

		//TODO code pertaining to reference Sequence is not added; look into original

		if (showAnnotation) {
			x = 2 * displayProperties.getActivityWidth();
			List<LineAnnotation> lineAnnotationList = displayProperties.getVisibleLineAnnotations(sequence);
			int k = 1;
			for (LineAnnotation lineAnnotation : lineAnnotationList) {
				y = k * activityHeight;
				fontY = y + displayProperties.getFontMetrics().getMaxAscent();
				g.drawString(lineAnnotation.getName(), x, fontY);
				k++;
			}
		}
	}

	/**
	 * Double clickable sequence name
	 * 
	 * @param loc
	 */
	@SuppressWarnings("deprecation")
	public void edit(Point loc) {
//		System.out.println("In edit");

		//        Frame frame = JOptionPane.getFrameForComponent(this);
		EditSequenceNameDialog ed = new EditSequenceNameDialog(owner, displayProperties, sequence);
		ed.setLocation(loc);
		ed.show();
	}

	//SequenceListener interface implementation

	public void sequenceColorChanged(Sequence sequence) {
		if (this.sequence != sequence) {
			throw new RuntimeException("Bound to incorrect Sequence");
		}
		sequenceColor = sequence.getColor();
		if ((sequenceColor == null) || (sequenceColor == Color.white)) {
			sequenceColor = Color.black;
		}
		revalidate();
		repaint();
	}

	public void sequenceLineAnnotationsChanged(Sequence sequence) {
		if (this.sequence != sequence) {
			throw new RuntimeException("Bound to incorrect Sequence");
		}
		if (displayProperties.getAnnView(sequence)) {
			revalidate();
			repaint();
		}
	}

	public void sequenceNameChanged(Sequence sequence, String oldName) throws Exception {
		if (this.sequence != sequence) {
			throw new RuntimeException("Bound to incorrect Sequence");
		}
		revalidate();
		repaint();
	}

	public void sequenceActivityChanged(Sequence aaseq) {
		// Nothing to be done here
	}

	public void sequenceAnnotationChanged(Sequence sequence) {
		// Nothing to be done here
	}

	public void sequenceColumnAnnotationsChanged(Sequence sequence, int column) {
		// Nothing to be done here
	}

	//DisplayPropertiesListener interface implementation
	public void displayAnnViewChanged(DisplayProperties dp, Sequence sequence, boolean show) {
		if (dp != displayProperties) {
			throw new RuntimeException("Bound to incorrect DisplayProperties");
		}
		if (this.sequence == sequence) {
			revalidate();
			repaint();
		}
	}

	public void displayFontChanged(DisplayProperties dp) {
		if (dp != displayProperties) {
			throw new RuntimeException("Bound to incorrect DisplayProperties");
		}

		revalidate();
		repaint();
	}

	public void displayColorSchemeChanged(DisplayProperties dp, ColorScheme old) {
		// Nothing to be done here
	}

	public void displayGroupEditingChanged(DisplayProperties dp) {
		// Nothing to be done here
	}

	public void displayHighlightsChanged(DisplayProperties dp, Sequence sequence) {
		// Nothing to be done here
	}

	public void displayHighlightsChanged(DisplayProperties dp, Sequence[] sequenceArray) {
		// Nothing to be done here
	}

	public void displayOverwriteChanged(DisplayProperties dp) {
		// Nothing to be done here
	}

	public void displayRenderGapsChanged(DisplayProperties dp) {
		// Nothing to be done here
	}

	public void activityBackgroundChanged(DisplayProperties dp) {
		// Nothing to be done here
	}

	public void displaySeqSelectChanged(DisplayProperties dp, Sequence sequence, boolean select) {
		// Nothing to be done here
	}

}
