// Om Ganesayanamaha
package org.processmining.plugins.tracealignment.util;

import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import javax.swing.JFrame;

import org.processmining.plugins.tracealignment.swingx.ErrorDialog;

public class FeatureExtraction {

	class DescendingStringLengthComparator implements Comparator<String> {
		public int compare(String str1, String str2) {
			return (str1.length() > str2.length()) ? 1 : str1.compareTo(str2);
		}
	}

	public enum FeatureType {
		INDIVIDUAL_EVENT, KGRAM, LOOP, LOOP_ALPHABET, MAXIMAL_REPEAT, SUPERMAXIMAL_REPEAT, NEARSUPERMAXIMAL_REPEAT, MAXIMAL_REPEAT_ALPHABET, SUPERMAXIMAL_REPEAT_ALPHABET, NEARSUPERMAXIMAL_REPEAT_ALPHABET
	};

	/*
	 * Sequence based feature set; INDIVIDUAL_EVENT, KGRAM, LOOP and {MAXIMAL,
	 * SUPERMAXIMAL, NEARSUPERMAXIMAL} Repeat features constitute the sequence
	 * based features
	 */
	Set<String> sequenceFeatureSet;

	/*
	 * For the alphabet based features, apart from the alphabet, we also need
	 * the sequences under the equivalence class of the alphabet; so, we need a
	 * map of alphabet (expressed as a Set<String)) and its corresponding
	 * equivalence class of sequences (repeats) expressed as Set<String>
	 * 
	 * {LOOP, MAXIMAL_REPEAT, NEARSUPERMAXIMAL_REPEAT,
	 * SUPERMAXIMAL_REPEAT}_ALPHABET features constitute this variable
	 */
	Map<Set<String>, Set<String>> alphabetFeatureSetMap;

	int encodingLength;
	List<String> charStreamList;

	/*
	 * ActivityCharMap is required for the repeat based features to form the
	 * combined charStream; Each trace need to be delimited by a distinct symbol
	 * which is provided in the activityCharMap as Tracei
	 */
	Map<String, String> activityCharMap;

	public FeatureExtraction() {

	}

	public FeatureExtraction(int encodingLength, List<String> charStreamList, Map<String, String> activityCharMap,
			FeatureType featureType) {
		this.encodingLength = encodingLength;
		this.charStreamList = charStreamList;
		this.activityCharMap = activityCharMap;

		switch (featureType) {
			case INDIVIDUAL_EVENT :
				computeIndividualEventFeatureSet();
				break;
			case LOOP :
				computeLoopFeatureSet();
				break;
			case LOOP_ALPHABET :
				computeLoopFeatureSet();
				EquivalenceClass equivalenceClass = new EquivalenceClass();
				alphabetFeatureSetMap = equivalenceClass.getAlphabetEquivalenceClassMap(encodingLength,
						sequenceFeatureSet);
				break;
			case MAXIMAL_REPEAT :
			case SUPERMAXIMAL_REPEAT :
			case NEARSUPERMAXIMAL_REPEAT :
				computeRepeatFeatureSet(featureType);
				break;
			case MAXIMAL_REPEAT_ALPHABET :
			case SUPERMAXIMAL_REPEAT_ALPHABET :
			case NEARSUPERMAXIMAL_REPEAT_ALPHABET :
				computeRepeatAlphabetFeatureSet(featureType);
				break;
			case KGRAM :
				System.out
						.println("Gram Size not specified; Can't proceed; Use other constructor for FeatureExtraction");
				System.exit(0);
				break;
			default :
				break;
		}
	}

	public FeatureExtraction(int encodingLength, List<String> charStreamList, FeatureType featureType, int gramSize) {
		if (featureType == FeatureType.KGRAM) {
			sequenceFeatureSet = new HashSet<String>();
			/*
			 * To reduce the computational/space complexity, we first remove all
			 * duplicates in the charStreamList
			 */
			Set<String> charStreamSet = new HashSet<String>();
			charStreamSet.addAll(charStreamList);

			int charStreamLength;
			for (String charStream : charStreamSet) {
				charStreamLength = charStream.length() / encodingLength;
				for (int i = 0; i < charStreamLength - gramSize + 1; i++) {
					sequenceFeatureSet.add(charStream.substring(i * encodingLength, (i + gramSize) * encodingLength));
				}
			}
		} else {
			System.out
					.println("Wrong Constructor invoked for FeatureExtraction; This constructor is to be used only for KGRAM feature type");
			System.exit(0);
		}
	}

	private void computeIndividualEventFeatureSet() {
		sequenceFeatureSet = new HashSet<String>();

		/*
		 * To reduce the computational/space complexity, we first remove all
		 * duplicates in the charStreamList
		 */
		Set<String> charStreamSet = new HashSet<String>();
		for (String charStream : charStreamList) {
			charStreamSet.add(charStream);
		}

		/*
		 * Process each charStream to extract the feature (individual event) and
		 * add it to the sequenceFeatureSet
		 */
		int charStreamLength;
		for (String charStream : charStreamSet) {
			charStreamLength = charStream.length() / encodingLength;
			for (int i = 0; i < charStreamLength; i++) {
				sequenceFeatureSet.add(charStream.substring(i * encodingLength, (i + 1) * encodingLength));
			}
		}
	}

	private void computeLoopFeatureSet() {
		sequenceFeatureSet = new HashSet<String>();

		/*
		 * To reduce the computational/space complexity, we first remove all
		 * duplicates in the charStreamList
		 */
		Set<String> charStreamSet = new HashSet<String>();
		for (String charStream : charStreamList) {
			charStreamSet.add(charStream);
		}

		UkkonenSuffixTree suffixTree;
		Map<TreeSet<String>, TreeSet<String>> loopAlphabetLoopPatternSetMap;
		/*
		 * For each charStream compute the set of tandem repeats
		 */
		for (String charStream : charStreamSet) {
			suffixTree = new UkkonenSuffixTree(encodingLength, charStream);
			suffixTree.LZDecomposition();
			//TODO
			/*
			 * Currently, the ukkonenSuffixTree code, the simple/complex tandem
			 * repeats are computed in the getPrimitiveTandemRepeats method; we
			 * need to modify this when ukkonenSuffixTree is recoded
			 */
			loopAlphabetLoopPatternSetMap = suffixTree.getPrimitiveTandemRepeats();

			// Extract the loop patterns from the map and put it in the sequenceFeatureSet
			for (TreeSet<String> loopAlphabet : loopAlphabetLoopPatternSetMap.keySet()) {
				sequenceFeatureSet.addAll(loopAlphabetLoopPatternSetMap.get(loopAlphabet));
			}
			suffixTree = null;
		}

		//Set no longer required variables to null
		charStreamSet = null;
	}

	private void computeRepeatFeatureSet(FeatureType featureType) {
		sequenceFeatureSet = new HashSet<String>();

		/*
		 * To reduce the computational/space complexity, we first remove all
		 * duplicates in the charStreamList
		 */
		Set<String> charStreamSet = new HashSet<String>();
		for (String charStream : charStreamList) {
			charStreamSet.add(charStream);
		}

		StringBuilder combinedCharStreamBuilder = new StringBuilder();
		int charStreamIndex = 0;
		int totalCharStreamLength = 0;
		for (String charStream : charStreamSet) {
			combinedCharStreamBuilder.append(charStream);
			totalCharStreamLength += charStream.length();
			if (activityCharMap.containsKey("Trace" + charStreamIndex)) {
				combinedCharStreamBuilder.append(activityCharMap.get("Trace" + charStreamIndex));
			} else {
				ErrorDialog.showErrorDialog(new JFrame(),"Activity Char Map doesn't contain encoding for trace indices");
				return;
			}
			charStreamIndex++;
		}

		UkkonenSuffixTree suffixTree = new UkkonenSuffixTree(encodingLength, combinedCharStreamBuilder.toString());
		suffixTree.findLeftDiverseNodes();
		switch (featureType) {
			case MAXIMAL_REPEAT :
				sequenceFeatureSet.addAll(suffixTree.getMaximalRepeats());
				break;
			case SUPERMAXIMAL_REPEAT :
				sequenceFeatureSet.addAll(suffixTree.getSuperMaximalRepeats());
				break;
			case NEARSUPERMAXIMAL_REPEAT :
				sequenceFeatureSet.addAll(suffixTree.getNearSuperMaximalRepeats());
				break;
			default :
				break;
		}

		//Set no longer required variables to null
		combinedCharStreamBuilder = null;
		charStreamSet = null;
		suffixTree = null;
	}

	private void computeRepeatAlphabetFeatureSet(FeatureType featureType) {
		alphabetFeatureSetMap = new HashMap<Set<String>, Set<String>>();

		/*
		 * To reduce the computational/space complexity, we first remove all
		 * duplicates in the charStreamList
		 */
		Set<String> charStreamSet = new HashSet<String>();
		for (String charStream : charStreamList) {
			charStreamSet.add(charStream);
		}

		StringBuilder combinedCharStreamBuilder = new StringBuilder();
		int charStreamIndex = 0;
		for (String charStream : charStreamSet) {
			combinedCharStreamBuilder.append(charStream);
			if (activityCharMap.containsKey("Trace" + charStreamIndex)) {
				combinedCharStreamBuilder.append(activityCharMap.get("Trace" + charStreamIndex));
			} else {
				ErrorDialog.showErrorDialog(new JFrame(),"Activity Char Map doesn't contain encoding for trace indices");
				return;
			}
			charStreamIndex++;
		}

		UkkonenSuffixTree suffixTree = new UkkonenSuffixTree(encodingLength, combinedCharStreamBuilder.toString());
		suffixTree.findLeftDiverseNodes();
		EquivalenceClass equivalenceClass = new EquivalenceClass();
		switch (featureType) {
			case MAXIMAL_REPEAT_ALPHABET :
				alphabetFeatureSetMap = equivalenceClass.getAlphabetEquivalenceClassMap(encodingLength, suffixTree
						.getMaximalRepeats());
				break;
			case SUPERMAXIMAL_REPEAT_ALPHABET :
				alphabetFeatureSetMap = equivalenceClass.getAlphabetEquivalenceClassMap(encodingLength, suffixTree
						.getSuperMaximalRepeats());
				break;
			case NEARSUPERMAXIMAL_REPEAT_ALPHABET :
				alphabetFeatureSetMap = equivalenceClass.getAlphabetEquivalenceClassMap(encodingLength, suffixTree
						.getNearSuperMaximalRepeats());
				break;
			default :
				break;
		}
	}

	public Set<String> getSequenceFeatureSet() {
		return sequenceFeatureSet;
	}

	public Map<Set<String>, Set<String>> getAlphabetFeatureSetMap() {
		return alphabetFeatureSetMap;
	}

	public Map<FeatureType, Set<String>> getRepeatFeatureTypeFeatureSetMap(int encodingLength,
			List<String> charStreamList, Map<String, String> activityCharStreamMap, List<FeatureType> featureTypeList) {
		StringBuilder combinedStreamBuilder = new StringBuilder();

		Set<String> charStreamSet = new HashSet<String>();
		charStreamSet.addAll(charStreamList);
		int traceIndex = 0;
		for (String charStream : charStreamSet) {
			combinedStreamBuilder.append(charStream);
			if (activityCharStreamMap.containsKey("Trace" + traceIndex)) {
				combinedStreamBuilder.append(activityCharStreamMap.get("Trace" + traceIndex));
			} else {
				return null;
			}
		}
		return getRepeatFeatureTypeFeatureSetMap(encodingLength, combinedStreamBuilder.toString(), featureTypeList);
	}

	public Map<FeatureType, Set<String>> getRepeatFeatureTypeFeatureSetMap(int encodingLength, String charStream,
			List<FeatureType> featureTypeList) {
		Map<FeatureType, Set<String>> repeatFeatureTypeFeatureSetMap = new HashMap<FeatureType, Set<String>>();

		UkkonenSuffixTree suffixTree = new UkkonenSuffixTree(encodingLength, charStream);
		suffixTree.findLeftDiverseNodes();
		for (FeatureType featureType : featureTypeList) {
			switch (featureType) {
				case MAXIMAL_REPEAT :
					repeatFeatureTypeFeatureSetMap.put(FeatureType.MAXIMAL_REPEAT, suffixTree.getMaximalRepeats());
					break;
				case SUPERMAXIMAL_REPEAT :
					repeatFeatureTypeFeatureSetMap.put(FeatureType.SUPERMAXIMAL_REPEAT, suffixTree
							.getSuperMaximalRepeats());
					break;
				case NEARSUPERMAXIMAL_REPEAT :
					repeatFeatureTypeFeatureSetMap.put(FeatureType.NEARSUPERMAXIMAL_REPEAT, suffixTree
							.getNearSuperMaximalRepeats());
					break;
				default :
					return null;
			}
		}

		return repeatFeatureTypeFeatureSetMap;
	}

	public Map<FeatureType, Map<Set<String>, Set<String>>> getRepeatAlphabetFeatureTypeFeatureSetMap(
			int encodingLength, List<String> charStreamList, Map<String, String> activityCharStreamMap,
			List<FeatureType> featureTypeList) {
		StringBuilder combinedStreamBuilder = new StringBuilder();

		Set<String> charStreamSet = new HashSet<String>();
		charStreamSet.addAll(charStreamList);
		int traceIndex = 0;
		for (String charStream : charStreamSet) {
			combinedStreamBuilder.append(charStream);
			if (activityCharStreamMap.containsKey("Trace" + traceIndex)) {
				combinedStreamBuilder.append(activityCharStreamMap.get("Trace" + traceIndex));
			} else {
				return null;
			}
		}

		return getRepeatAlphabetFeatureTypeFeatureSetMap(encodingLength, combinedStreamBuilder.toString(),
				featureTypeList);
	}

	public Map<FeatureType, Map<Set<String>, Set<String>>> getRepeatAlphabetFeatureTypeFeatureSetMap(
			int encodingLength, String charStream, List<FeatureType> featureTypeList) {
		Map<FeatureType, Map<Set<String>, Set<String>>> repeatAlphabetFeatureTypeFeatureSetMap = new HashMap<FeatureType, Map<Set<String>, Set<String>>>();

		UkkonenSuffixTree suffixTree = new UkkonenSuffixTree(encodingLength, charStream);
		suffixTree.findLeftDiverseNodes();
		EquivalenceClass equivalenceClass = new EquivalenceClass();
		for (FeatureType featureType : featureTypeList) {
			switch (featureType) {
				case MAXIMAL_REPEAT_ALPHABET :
					repeatAlphabetFeatureTypeFeatureSetMap.put(FeatureType.MAXIMAL_REPEAT_ALPHABET, equivalenceClass
							.getAlphabetEquivalenceClassMap(encodingLength, suffixTree.getMaximalRepeats()));
					break;
				case SUPERMAXIMAL_REPEAT_ALPHABET :
					repeatAlphabetFeatureTypeFeatureSetMap.put(FeatureType.SUPERMAXIMAL_REPEAT_ALPHABET,
							equivalenceClass.getAlphabetEquivalenceClassMap(encodingLength, suffixTree
									.getSuperMaximalRepeats()));
					break;
				case NEARSUPERMAXIMAL_REPEAT_ALPHABET :
					repeatAlphabetFeatureTypeFeatureSetMap.put(FeatureType.NEARSUPERMAXIMAL_REPEAT, equivalenceClass
							.getAlphabetEquivalenceClassMap(encodingLength, suffixTree.getNearSuperMaximalRepeats()));
					break;
				default :
					return null;
			}
		}

		return repeatAlphabetFeatureTypeFeatureSetMap;
	}
}
